@@ -1,6 +1,6 @@
 /* -*-c++-*- */
 /* osgEarth - Dynamic map generation toolkit for OpenSceneGraph
- * Copyright 2015 Pelican Mapping
+ * Copyright 2016 Pelican Mapping
  * http://osgearth.org
  *
  * osgEarth is free software; you can redistribute it and/or modify
@@ -34,6 +34,12 @@
 #include <algorithm>
 #include <curl/curl.h>
 
+// Whether to use WinInet instead of cURL - CMAKE option
+#ifdef OSGEARTH_USE_WININET_FOR_HTTP
+#include <WinInet.h>
+#pragma comment(lib, "wininet.lib")
+#endif
+
 #define LC "[HTTPClient] "
 
 //#define OE_TEST OE_NOTICE
@@ -263,9 +269,11 @@ HTTPRequest::getURL() const
 
 /****************************************************************************/
 
-HTTPResponse::HTTPResponse( long _code )
-: _response_code( _code ),
-  _cancelled(false)
+HTTPResponse::HTTPResponse( long _code ) :
+_response_code( _code ),
+_cancelled(false),
+_duration_s(0.0),
+_lastModified(0u)
 {
     _parts.reserve(1);
 }
@@ -274,7 +282,9 @@ HTTPResponse::HTTPResponse( const HTTPResponse& rhs ) :
 _response_code( rhs._response_code ),
 _parts( rhs._parts ),
 _mimeType( rhs._mimeType ),
-_cancelled( rhs._cancelled )
+_cancelled( rhs._cancelled ),
+_duration_s(0.0),
+_lastModified(0u)
 {
     //nop
 }
@@ -316,8 +326,8 @@ HTTPResponse::getPartStream( unsigned int n ) const {
 
 std::string
 HTTPResponse::getPartAsString( unsigned int n ) const {
-     std::string streamStr;
-     streamStr = _parts[n]->_stream.str();
+    std::string streamStr;
+    streamStr = _parts[n]->_stream.str();    
     return streamStr;
 }
 
@@ -381,7 +391,8 @@ HTTPClient::getClient()
 HTTPClient::HTTPClient() :
 _initialized    ( false ),
 _curl_handle    ( 0L ),
-_simResponseCode( -1L )
+_simResponseCode( -1L ),
+_previousHttpAuthentication(0L)
 {
     //nop
     //do no CURL calls here.
@@ -418,6 +429,13 @@ HTTPClient::initializeImpl()
         OE_WARN << LC << "Simulating a network error with Response Code = " << _simResponseCode << std::endl;
     }
 
+    // Check to HTTP disabling (for testing)
+    const char* disable = getenv("OSGEARTH_HTTP_DISABLE");
+    if (disable)
+    {
+        _simResponseCode = 503L; // SERVICE UNAVAILABLE
+    }
+
     // Dumps out HTTP request/response info
     if ( ::getenv("OSGEARTH_HTTP_DEBUG") )
     {
@@ -436,6 +454,10 @@ HTTPClient::initializeImpl()
     curl_easy_setopt( _curl_handle, CURLOPT_NOPROGRESS, (void*)0 ); //0=enable.
     curl_easy_setopt( _curl_handle, CURLOPT_FILETIME, true );
 
+    // Enable automatic CURL decompression of known types. An empty string will automatically add all supported encoding types that are built into curl.
+    // Note that you must have curl built against zlib to support gzip or deflate encoding.
+    curl_easy_setopt( _curl_handle, CURLOPT_ENCODING, "");
+
     osg::ref_ptr< CurlConfigHandler > curlConfigHandler = getCurlConfigHandler();
     if (curlConfigHandler.valid()) {
         curlConfigHandler->onInitialize(_curl_handle);
@@ -695,6 +717,246 @@ HTTPClient::download(const std::string& uri,
     return getClient().doDownload( uri, localPath );
 }
 
+
+#ifdef OSGEARTH_USE_WININET_FOR_HTTP
+
+namespace
+{
+    std::string GetLastErrorAsString()
+    {
+        //Get the error message, if any.
+        DWORD errorMessageID = ::GetLastError();
+        if(errorMessageID == 0)
+            return std::string("Error Code 0"); //No error message has been recorded
+
+        LPSTR messageBuffer = nullptr;
+        size_t size = FormatMessageA(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
+                                     NULL, errorMessageID, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPSTR)&messageBuffer, 0, NULL);
+
+        std::string message(messageBuffer, size);
+
+        //Free the buffer.
+        LocalFree(messageBuffer);
+
+        message = Stringify() << "[Code " << errorMessageID << "] " << message;
+
+        return message;
+    }
+}
+
+HTTPResponse
+HTTPClient::doGet(const HTTPRequest&    request,
+                  const osgDB::Options* options, 
+                  ProgressCallback*     progress) const
+{
+    OE_START_TIMER(http_get);
+
+    std::string url = request.getURL();
+    // Rewrite the url if the url rewriter is available  
+    osg::ref_ptr< URLRewriter > rewriter = getURLRewriter();
+    if ( rewriter.valid() )
+    {
+        std::string oldURL = url;
+        url = rewriter->rewrite( oldURL );
+        OE_DEBUG << LC << "Rewrote URL " << oldURL << " to " << url << std::endl;
+    }
+
+    HINTERNET hInternet = InternetOpen(
+        getUserAgent().c_str(),
+        //"Mozilla/5.0 (Windows NT 6.1; WOW64; Trident/7.0; AS; rv:11.0) like Gecko",
+        INTERNET_OPEN_TYPE_PRECONFIG, 
+        NULL,       // proxy
+        NULL,       // proxy bypass
+        0);         // flags
+
+    if( !hInternet )
+    {
+        OE_WARN << LC << "InternetOpen failed: " << GetLastErrorAsString() << std::endl;
+        return HTTPResponse(0);
+    }
+
+    // clears any session cookies..?
+    InternetSetOption( 0, INTERNET_OPTION_END_BROWSER_SESSION, NULL, 0 );
+    
+    // parse the URL:
+    URL_COMPONENTS urlcomp;
+    ZeroMemory(&urlcomp, sizeof(urlcomp));
+	urlcomp.dwStructSize = sizeof(urlcomp);
+	urlcomp.dwHostNameLength = 1;
+    urlcomp.dwUserNameLength = 1;
+    urlcomp.dwPasswordLength = 1;
+    urlcomp.dwUrlPathLength  = 1;
+
+    if ( !InternetCrackUrl(url.c_str(), 0, 0L, &urlcomp) )
+    {
+        OE_WARN << LC << "InternetCrackUrl failed for " << url << ": " << GetLastErrorAsString() << std::endl;
+        InternetCloseHandle( hInternet );
+        return HTTPResponse(0);
+    }
+    
+    WORD port =
+        urlcomp.nPort != 0 ? urlcomp.nPort :
+        urlcomp.nScheme == INTERNET_SCHEME_HTTPS ? INTERNET_DEFAULT_HTTPS_PORT :
+        INTERNET_DEFAULT_HTTP_PORT;
+
+    std::string hostName( urlcomp.lpszHostName );
+    int slash = hostName.find_first_of('/');
+    if ( slash != std::string::npos )
+        hostName = hostName.substr(0, slash);
+
+    OE_DEBUG
+        << "\n"
+        << "Host name = " << hostName << "\n"
+        << "Url path = " << urlcomp.lpszUrlPath << "\n"
+        << "Port = " << port << "\n";
+    
+    DWORD openFlags =
+        //INTERNET_FLAG_PRAGMA_NOCACHE |
+        INTERNET_FLAG_RELOAD |
+        INTERNET_FLAG_NO_CACHE_WRITE |
+        INTERNET_FLAG_KEEP_CONNECTION;
+
+    if ( urlcomp.nScheme == INTERNET_SCHEME_HTTPS)
+    {
+        openFlags |= INTERNET_FLAG_SECURE;
+    }
+
+    // InternetOpenUrl is a lot less code, but we have to use the InternetConnnect +
+    // HttpOpenRequest + HttpSendRequest approach in order to support system dialogs
+    // for PKI certificates and username/password queries.
+    
+    HINTERNET hConnection = InternetConnect(
+        hInternet,
+        hostName.c_str(),
+        port,
+        "", // username
+        "", // password
+        INTERNET_SERVICE_HTTP,
+        0,  // flags
+        INTERNET_NO_CALLBACK); // context
+
+    if ( !hConnection )
+    {
+        OE_WARN << LC << "InternetConnect failed for " << url << ": " << GetLastErrorAsString() << std::endl;
+        InternetCloseHandle( hInternet );
+        return HTTPResponse(0);
+    }
+
+    HINTERNET hRequest = HttpOpenRequest(
+        hConnection,            // handle from InternetConnect
+        "GET",                  // verb
+        urlcomp.lpszUrlPath,    // path
+        NULL,                   // HTTP version (NULL = default)
+        NULL,                   // Referrer
+        NULL,                   // Accept types
+        openFlags,              // flags
+        INTERNET_NO_CALLBACK);                  // context (user data)
+
+    if ( !hRequest )
+    {
+        OE_WARN << LC << "HttpOpenRequest failed for " << url << ": " << GetLastErrorAsString() << std::endl;
+        InternetCloseHandle( hConnection );
+        InternetCloseHandle( hInternet );
+        return HTTPResponse(0);
+    }
+
+    while( !HttpSendRequest(hRequest, NULL, 0, NULL, 0) )
+    {
+        DWORD errorNum = GetLastError();
+            
+        // Request for client cert; open system dialog.
+        if ( errorNum == ERROR_INTERNET_CLIENT_AUTH_CERT_NEEDED )
+        {
+            OE_WARN << LC << "Server reports ERROR_INTERNET_CLIENT_AUTH_CERT_NEEDED!\n";
+
+            // Return ERROR_SUCCESS regardless of clicking on OK or Cancel
+            DWORD dialogResult = InternetErrorDlg(
+                GetDesktopWindow(), 
+                hRequest,
+                ERROR_INTERNET_CLIENT_AUTH_CERT_NEEDED,
+                FLAGS_ERROR_UI_FILTER_FOR_ERRORS | FLAGS_ERROR_UI_FLAGS_GENERATE_DATA | FLAGS_ERROR_UI_FLAGS_CHANGE_OPTIONS, 
+                NULL );
+
+            if ( dialogResult != ERROR_SUCCESS )
+            {
+                OE_WARN << LC << "InternetErrorDlg failed to produce client cert " << url << ": " << GetLastErrorAsString() << std::endl;
+                InternetCloseHandle( hRequest );
+                InternetCloseHandle( hConnection );
+                InternetCloseHandle( hInternet );
+                return HTTPResponse(0);
+            }
+        }
+
+        else
+        {
+            OE_WARN << LC << "HttpSendRequest failed to open " << url << ": " << GetLastErrorAsString() << std::endl;
+            InternetCloseHandle( hRequest );
+            InternetCloseHandle( hConnection );
+            InternetCloseHandle( hInternet );
+            return HTTPResponse(0);
+        }
+    }
+
+    int statusCode = 0;
+    std::string contentType;
+
+    char  buffer[4096];
+    DWORD bufferLen = 4096;
+    DWORD index = 0;
+
+    if ( HttpQueryInfo(hRequest, HTTP_QUERY_STATUS_CODE, buffer, &bufferLen, &index) )
+    {             
+        statusCode = as<int>( std::string(buffer, bufferLen), 0 );
+    }
+    
+    HTTPResponse response(statusCode);
+
+    bufferLen = 4096, index = 0;
+    if ( HttpQueryInfo(hRequest, HTTP_QUERY_CONTENT_TYPE, buffer, &bufferLen, &index) )
+    {
+        response._mimeType = std::string(buffer, bufferLen);
+    }
+
+    bufferLen = 4096, index = 0;
+    if ( HttpQueryInfo(hRequest, HTTP_QUERY_LAST_MODIFIED, buffer, &bufferLen, &index) )
+    {
+        response._lastModified = as<long>(std::string(buffer, bufferLen), 0L);
+    }
+
+    response._mimeType = contentType;
+
+    if ( statusCode == 200 )
+    {    
+        osg::ref_ptr<HTTPResponse::Part> part = new HTTPResponse::Part();
+
+        DWORD numBytesRead = 0;
+        while( InternetReadFile(hRequest, buffer, 4096, &numBytesRead) && numBytesRead )
+        {
+            part->_stream << std::string(buffer, numBytesRead);
+        }
+
+        response._parts.push_back( part.get() );
+    }
+
+    InternetCloseHandle( hRequest );
+    InternetCloseHandle( hConnection );
+    InternetCloseHandle( hInternet );
+
+    response._duration_s = OE_STOP_TIMER(http_get);
+
+    if ( progress )
+    {
+        progress->stats("http_get_time") += OE_GET_TIMER(http_get);
+        progress->stats("http_get_count") += 1;
+        if ( response._cancelled )
+            progress->stats("http_cancel_count") += 1;
+    }
+    
+    return response;
+}
+
+#else // OSGEARTH_USE_WININET_FOR_HTTP
+
 HTTPResponse
 HTTPClient::doGet(const HTTPRequest&    request,
                   const osgDB::Options* options, 
@@ -703,6 +965,8 @@ HTTPClient::doGet(const HTTPRequest&    request,
     initialize();
 
     OE_START_TIMER(http_get);
+    
+    std::string url = request.getURL();
 
     const osgDB::AuthenticationMap* authenticationMap = (options && options->getAuthenticationMap()) ? 
             options->getAuthenticationMap() :
@@ -798,14 +1062,13 @@ HTTPClient::doGet(const HTTPRequest&    request,
         curl_easy_setopt( _curl_handle, CURLOPT_PROXY, 0 );
     }
 
-    std::string url = request.getURL();
     // Rewrite the url if the url rewriter is available  
     osg::ref_ptr< URLRewriter > rewriter = getURLRewriter();
     if ( rewriter.valid() )
     {
         std::string oldURL = url;
         url = rewriter->rewrite( oldURL );
-        OE_INFO << LC << "Rewrote URL " << oldURL << " to " << url << std::endl;
+        OE_DEBUG << LC << "Rewrote URL " << oldURL << " to " << url << std::endl;
     }
 
     const osgDB::AuthenticationDetails* details = authenticationMap ?
@@ -846,7 +1109,7 @@ HTTPClient::doGet(const HTTPRequest&    request,
             const_cast<HTTPClient*>(this)->_previousHttpAuthentication = 0;
         }
 #endif
-    }
+    }  
 
 
     // Set any headers
@@ -859,11 +1122,11 @@ HTTPClient::doGet(const HTTPRequest&    request,
             buf << itr->first << ": " << itr->second;
             headers = curl_slist_append(headers, buf.str().c_str());
         }
-    }    
+    }
 
     // Disable the default Pragma: no-cache that curl adds by default.
     headers = curl_slist_append(headers, "Pragma: ");
-    curl_easy_setopt(_curl_handle, CURLOPT_HTTPHEADER, headers);
+    curl_easy_setopt(_curl_handle, CURLOPT_HTTPHEADER, headers); 
     
     osg::ref_ptr<HTTPResponse::Part> part = new HTTPResponse::Part();
     StreamObject sp( &part->_stream );
@@ -933,6 +1196,9 @@ HTTPClient::doGet(const HTTPRequest&    request,
         response._mimeType = content_type_cp;    
     } 
 
+    // read the file time:
+    response._lastModified = getCurlFileTime( _curl_handle );
+
     // upon success, parse the data:
     if ( res != CURLE_ABORTED_BY_CALLBACK && res != CURLE_OPERATION_TIMEDOUT )
     {        
@@ -1035,6 +1301,8 @@ HTTPClient::doGet(const HTTPRequest&    request,
     return response;
 }
 
+#endif // USE_WININET
+
 bool
 HTTPClient::doDownload(const std::string& url, const std::string& filename)
 {
@@ -1151,7 +1419,7 @@ HTTPClient::doReadImage(const HTTPRequest&    request,
         }
         
         // last-modified (file time)
-        result.setLastModifiedTime( getCurlFileTime(_curl_handle) );
+        result.setLastModifiedTime( response._lastModified ); //getCurlFileTime(_curl_handle) );
         
         // Time of query
         result.setDuration( response.getDuration() );
@@ -1230,7 +1498,7 @@ HTTPClient::doReadNode(const HTTPRequest&    request,
         }
         
         // last-modified (file time)
-        result.setLastModifiedTime( getCurlFileTime(_curl_handle) );
+        result.setLastModifiedTime( response._lastModified ); //getCurlFileTime(_curl_handle) );
     }
     else
     {
@@ -1302,7 +1570,7 @@ HTTPClient::doReadObject(const HTTPRequest&    request,
         }
         
         // last-modified (file time)
-        result.setLastModifiedTime( getCurlFileTime(_curl_handle) );
+        result.setLastModifiedTime( response._lastModified ); //getCurlFileTime(_curl_handle) );
     }
     else
     {
@@ -1384,7 +1652,7 @@ HTTPClient::doReadString(const HTTPRequest&    request,
     result.setMetadata( response.getHeadersAsConfig() );
 
     // last-modified (file time)
-    result.setLastModifiedTime( getCurlFileTime(_curl_handle) );
+    result.setLastModifiedTime( response._lastModified ); //getCurlFileTime(_curl_handle) );
 
     return result;
 }
