@@ -1,6 +1,6 @@
 /* -*-c++-*- */
 /* osgEarth - Dynamic map generation toolkit for OpenSceneGraph
- * Copyright 2015 Pelican Mapping
+ * Copyright 2016 Pelican Mapping
  * http://osgearth.org
  *
  * osgEarth is free software; you can redistribute it and/or modify
@@ -69,25 +69,35 @@ _dataModelRevision   ( 0 )
             << _mapOptions.cachePolicy()->usageString() << ")" << std::endl;
     }
 
-    // Combine the CachePolicy in the map options with the settings in
-    // the registry.
-    Registry::instance()->resolveCachePolicy( _mapOptions.cachePolicy() );
-
     // the map-side dbOptions object holds I/O information for all components.
-    _dbOptions = osg::clone( Registry::instance()->getDefaultOptions() );
+    _readOptions = osg::clone( Registry::instance()->getDefaultOptions() );
+
+    // put the CacheSettings object in there. We will propogate this throughout
+    // the data model and the renderer.
+    CacheSettings* cacheSettings = new CacheSettings();
+
+    // Set up a cache if there's one in the options:
+    if (_mapOptions.cache().isSet())
+        cacheSettings->setCache(CacheFactory::create(_mapOptions.cache().get()));
+
+    // Otherwise use the registry default cache if there is one:
+    if (cacheSettings->getCache() == 0L)
+        cacheSettings->setCache(Registry::instance()->getDefaultCache());
+
+    // Integrate local cache policy (which can be overridden by the environment)
+    cacheSettings->integrateCachePolicy(_mapOptions.cachePolicy());
+
+    // store in the options so we can propagate it to layers, etc.
+    cacheSettings->store(_readOptions.get());
+
+    OE_INFO << LC << cacheSettings->toString() << "\n";
+
+
+    // remember the referrer for relative-path resolution:
+    URIContext( _mapOptions.referrer() ).store( _readOptions.get() );
 
     // we do our own caching
-    _dbOptions->setObjectCacheHint( osgDB::Options::CACHE_NONE );
-
-    // store the IO information in the top-level DB Options:
-    _mapOptions.cachePolicy()->apply( _dbOptions.get() );
-    URIContext( _mapOptions.referrer() ).apply( _dbOptions.get() );
-
-    // apply an express tile size if there is one.
-    if ( _mapOptions.elevationTileSize().isSet() )
-    {
-        _elevationLayers.setExpressTileSize( *_mapOptions.elevationTileSize() );
-    }
+    _readOptions->setObjectCacheHint( osgDB::Options::CACHE_NONE );
 
     // set up a callback that the Map will use to detect Elevation Layer
     // visibility changes
@@ -305,8 +315,6 @@ Map::setName( const std::string& name ) {
 Revision
 Map::getDataModelRevision() const
 {
-    //Don't really need this here.
-    //Threading::ScopedReadLock lock( const_cast<Map*>(this)->_mapDataMutex );
     return _dataModelRevision;
 }
 
@@ -321,53 +329,17 @@ Map::getProfile() const
 Cache*
 Map::getCache() const
 {
-    if ( !_cache.valid() )
-    {
-        Cache* cache = 0L;
-        
-        // if there's a cache in the registry, install it now.
-        if ( Registry::instance()->getCache() )
-        {
-            cache = Registry::instance()->getCache();
-        }
-
-        // or, if a cache is defined in the options, use that.
-        else if ( _mapOptions.cache().isSet() )
-        {
-            cache = CacheFactory::create( _mapOptions.cache().get() );
-        }
-
-        if ( cache )
-        {
-            const_cast<Map*>(this)->setCache( cache );
-        }
-    }
-    return _cache.get();
+    CacheSettings* cacheSettings = CacheSettings::get(_readOptions.get());
+    return cacheSettings ? cacheSettings->getCache() : 0L;
 }
 
 void
-Map::setCache( Cache* cache )
+Map::setCache(Cache* cache)
 {
-    if (_cache.get() != cache)
-    {
-        _cache = cache;
-
-        if ( _cache.valid() )
-        {
-            _cache->apply( _dbOptions.get() );
-        }
-
-        // Propagate the cache to any of our layers
-        for (ImageLayerVector::iterator i = _imageLayers.begin(); i != _imageLayers.end(); ++i)
-        {
-            i->get()->setDBOptions( _dbOptions.get() );
-        }
-
-        for (ElevationLayerVector::iterator i = _elevationLayers.begin(); i != _elevationLayers.end(); ++i)
-        {
-            i->get()->setDBOptions( _dbOptions.get() );
-        }
-    }
+    // note- probably unsafe to do this after initializing the terrain. so don't.
+    CacheSettings* cacheSettings = CacheSettings::get(_readOptions.get());
+    if (cacheSettings && cacheSettings->getCache() != cache)
+        cacheSettings->setCache(cache);
 }
 
 void 
@@ -417,10 +389,7 @@ Map::addImageLayer( ImageLayer* layer )
     if ( layer )
     {
         // Set the DB options for the map from the layer, including the cache policy.
-        layer->setDBOptions( _dbOptions.get() );
-
-        // propagate the cache to the layer:
-        layer->setCache( this->getCache() );
+        layer->setReadOptions( _readOptions.get() );
 
         // Tell the layer the map profile, if possible:
         if ( _profile.valid() )
@@ -428,6 +397,9 @@ Map::addImageLayer( ImageLayer* layer )
             layer->setTargetProfileHint( _profile.get() );
         }
 
+        // open the layer:
+        layer->open();
+
         int newRevision;
 
         // Add the layer to our stack.
@@ -456,14 +428,16 @@ Map::insertImageLayer( ImageLayer* layer, unsigned int index )
     if ( layer )
     {
         //Set options for the map from the layer
-        layer->setDBOptions( _dbOptions.get() );
-
-        //Set the Cache for the MapLayer to our cache.
-        layer->setCache( this->getCache() );
+        layer->setReadOptions( _readOptions.get() );
 
         // Tell the layer the map profile, if possible:
         if ( _profile.valid() )
+        {
             layer->setTargetProfileHint( _profile.get() );
+        }
+
+        // open the layer:
+        layer->open();
 
         int newRevision;
 
@@ -485,7 +459,7 @@ Map::insertImageLayer( ImageLayer* layer, unsigned int index )
             i->get()->onMapModelChanged( MapModelChange(
                 MapModelChange::ADD_IMAGE_LAYER, newRevision, layer, index) );
         }   
-    }   
+    } 
 }
 
 void
@@ -496,15 +470,14 @@ Map::addElevationLayer( ElevationLayer* layer )
     if ( layer )
     {
         //Set options for the map from the layer
-        layer->setDBOptions( _dbOptions.get() );
+        layer->setReadOptions( _readOptions.get() );
 
-        //Set the Cache for the MapLayer to our cache.
-        layer->setCache( this->getCache() );
-        
         // Tell the layer the map profile, if possible:
         if ( _profile.valid() )
             layer->setTargetProfileHint( _profile.get() );
 
+        layer->open();
+
         int newRevision;
 
         // Add the layer to our stack.
@@ -711,7 +684,10 @@ Map::addModelLayer( ModelLayer* layer )
         }
 
         // initialize the model layer
-        layer->initialize( _dbOptions.get() );
+        layer->setReadOptions(_readOptions.get());
+
+        // open it and check the status
+        layer->open();
 
         // a seprate block b/c we don't need the mutex
         for( MapCallbackList::iterator i = _mapCallbacks.begin(); i != _mapCallbacks.end(); i++ )
@@ -735,7 +711,10 @@ Map::insertModelLayer( ModelLayer* layer, unsigned int index )
         }
 
         // initialize the model layer
-        layer->initialize( _dbOptions.get() );
+        layer->setReadOptions(_readOptions.get());
+
+        // open and get the status
+        layer->open();
 
         // a seprate block b/c we don't need the mutex
         for( MapCallbackList::iterator i = _mapCallbacks.begin(); i != _mapCallbacks.end(); i++ )
@@ -835,7 +814,9 @@ Map::addTerrainMaskLayer( MaskLayer* layer )
             newRevision = ++_dataModelRevision;
         }
 
-        layer->initialize( _dbOptions.get(), this );
+        layer->setReadOptions(_readOptions.get());
+
+        layer->open();
 
         // a separate block b/c we don't need the mutex   
         for( MapCallbackList::iterator i = _mapCallbacks.begin(); i != _mapCallbacks.end(); i++ )
@@ -1108,42 +1089,6 @@ Map::calculateProfile()
     }
 }
 
-osg::HeightField*
-Map::createReferenceHeightField(const TileKey& key,
-                                bool           expressHeightsAsHAE) const
-{
-    unsigned size = std::max(*_mapOptions.elevationTileSize(), 2u);
-    return HeightFieldUtils::createReferenceHeightField(key.getExtent(), size, size, expressHeightsAsHAE);
-}
-
-#if 0
-bool
-Map::populateHeightField(osg::ref_ptr<osg::HeightField>& hf,
-                         const TileKey&                  key,
-                         bool                            convertToHAE,
-                         ElevationSamplePolicy           samplePolicy, // deprecated (unused)
-                         bool                            fallbackIfPossible,
-                         ProgressCallback*               progress) const
-{
-    Threading::ScopedReadLock lock( const_cast<Map*>(this)->_mapDataMutex );
-
-    ElevationInterpolation interp = getMapOptions().elevationInterpolation().get();    
-
-    if ( !hf.valid() )
-    {
-        hf = createReferenceHeightField(key, convertToHAE);
-    }
-
-    return _elevationLayers.populateHeightField(
-        hf.get(),
-        key,
-        convertToHAE ? _profileNoVDatum.get() : 0L,
-        interp,
-        fallbackIfPossible,
-        progress );
-}
-#endif
-
 const SpatialReference*
 Map::getWorldSRS() const
 {
@@ -1171,8 +1116,6 @@ Map::sync( MapFrame& frame ) const
         if ( frame._parts & ELEVATION_LAYERS )
         {
             frame._elevationLayers = _elevationLayers;
-            if ( _mapOptions.elevationTileSize().isSet() )
-                frame._elevationLayers.setExpressTileSize( *_mapOptions.elevationTileSize() );
         }
 
         if ( frame._parts & MODEL_LAYERS )
