@@ -1,6 +1,6 @@
 /* -*-c++-*- */
 /* osgEarth - Dynamic map generation toolkit for OpenSceneGraph
-* Copyright 2015 Pelican Mapping
+* Copyright 2016 Pelican Mapping
 * http://osgearth.org
 *
 * osgEarth is free software; you can redistribute it and/or modify
@@ -37,6 +37,8 @@
 #include <osgEarth/TerrainEngineNode>
 #include <osgEarth/TextureCompositor>
 #include <osgEarth/ShaderGenerator>
+#include <osgEarth/SpatialReference>
+#include <osgEarth/MapModelChange>
 #include <osgEarth/URI>
 #include <osg/ArgumentParser>
 #include <osg/PagedLOD>
@@ -47,7 +49,7 @@ using namespace osgEarth;
 
 //---------------------------------------------------------------------------
 
-namespace 
+namespace
 {
     // adapter that lets MapNode listen to Map events
     struct MapNodeMapCallbackProxy : public MapCallback
@@ -100,13 +102,13 @@ public:
       }
 
       virtual void apply(osg::PagedLOD& node)
-      {        
-          //The PagedLOD node will contain two filenames, the first is empty and is the actual geometry of the          
+      {
+          //The PagedLOD node will contain two filenames, the first is empty and is the actual geometry of the
           //tile and the second is the filename of the next tile.
           if (node.getNumFileNames() > 1)
           {
               //Get the child filename
-              const std::string &filename = node.getFileName(1);              
+              const std::string &filename = node.getFileName(1);
               if (osgEarth::Registry::instance()->isBlacklisted(filename))
               {
                   //If the tile is blacklisted, we set the actual geometry, child 0, to always display
@@ -125,16 +127,16 @@ public:
                       if (node.getRangeMode() == osg::LOD::PIXEL_SIZE_ON_SCREEN)
                       {
                           node.setRange( 0, ranges->_minRange, ranges->_maxRange );
-                          node.setRange( 1, ranges->_maxRange, FLT_MAX );                          
+                          node.setRange( 1, ranges->_maxRange, FLT_MAX );
                       }
                       else
-                      {                          
+                      {
                           node.setRange(0, ranges->_minRange, ranges->_maxRange);
                           node.setRange(1, 0, ranges->_minRange);
                       }
-                  }                  
+                  }
               }
-              
+
           }
           traverse(node);
       }
@@ -157,7 +159,27 @@ MapNode::load(osg::ArgumentParser& args)
                 return r.release<MapNode>();
             }
         }
-    }    
+    }
+    return 0L;
+}
+
+MapNode*
+MapNode::load(osg::ArgumentParser& args, const MapNodeOptions& defaults)
+{
+    for( int i=1; i<args.argc(); ++i )
+    {
+        if ( args[i] && endsWith(args[i], ".earth") )
+        {
+            osg::ref_ptr<osgDB::Options> dbo = new osgDB::Options();
+            std::string optionsJSON = defaults.getConfig().toJSON();
+            dbo->setPluginStringData( "osgEarth.defaultOptions", optionsJSON );
+            ReadResult r = URI(args[i]).readNode( dbo.get() );
+            if ( r.succeeded() )
+            {
+                return r.release<MapNode>();
+            }
+        }
+    }
     return 0L;
 }
 
@@ -207,7 +229,7 @@ MapNode::init()
     // deleting during startup. It is possible that during startup, a driver
     // will load that will take a reference to the MapNode (like in a
     // ModelSource node operation) and we don't want that deleting the MapNode
-    // out from under us. 
+    // out from under us.
     // This is paired by an unref_nodelete() at the end of this method.
     this->ref();
 
@@ -238,14 +260,14 @@ MapNode::init()
     // will make their way to any read* calls down the pipe
     const osgDB::Options* global_options = _map->getGlobalOptions();
 
-    osg::ref_ptr<osgDB::Options> local_options = global_options ? 
+    osg::ref_ptr<osgDB::Options> local_options = global_options ?
         Registry::instance()->cloneOrCreateOptions( global_options ) :
         NULL;
 
     if ( local_options.valid() )
     {
         OE_INFO << LC
-            << "Options string = " 
+            << "Options string = "
             << (local_options.valid()? local_options->getOptionString() : "<empty>")
             << std::endl;
     }
@@ -272,8 +294,8 @@ MapNode::init()
         _terrainEngineContainer->getOrCreateStateSet()->addUniform(
             Registry::shaderFactory()->createUniformForGLMode(GL_LIGHTING, *terrainOptions.enableLighting()) );
 
-        _terrainEngineContainer->getOrCreateStateSet()->setMode( 
-            GL_LIGHTING, 
+        _terrainEngineContainer->getOrCreateStateSet()->setMode(
+            GL_LIGHTING,
             terrainOptions.enableLighting().value() ? 1 : 0 );
     }
 
@@ -281,7 +303,7 @@ MapNode::init()
     {
         // inform the terrain engine of the map information now so that it can properly
         // initialize it's CoordinateSystemNode. This is necessary in order to support
-        // manipulators and to set up the texture compositor prior to frame-loop 
+        // manipulators and to set up the texture compositor prior to frame-loop
         // initialization.
         _terrainEngine->preInitialize( _map.get(), terrainOptions );
         _terrainEngineContainer->addChild( _terrainEngine );
@@ -297,7 +319,7 @@ MapNode::init()
     // model layer will still work OK though.
     _models = new osg::Group();
     _models->setName( "osgEarth::MapNode.modelsGroup" );
-    _models->getOrCreateStateSet()->setRenderBinDetails(2, "RenderBin");
+    //_models->getOrCreateStateSet()->setRenderBinDetails(1, "RenderBin");
     addChild( _models.get() );
 
     // a decorator for overlay models:
@@ -322,6 +344,7 @@ MapNode::init()
         if ( _mapNodeOptions.overlayResolutionRatio().isSet() )
             draping->setResolutionRatio( *_mapNodeOptions.overlayResolutionRatio() );
 
+        draping->reestablish( getTerrainEngine() );
         _overlayDecorator->addTechnique( draping );
     }
 
@@ -330,7 +353,6 @@ MapNode::init()
         _overlayDecorator->addTechnique( new ClampingTechnique() );
     }
 
-
     addTerrainDecorator( _overlayDecorator );
 
     // install any pre-existing model layers:
@@ -351,11 +373,11 @@ MapNode::init()
     if ( _mapNodeOptions.enableLighting().isSet() )
     {
         stateset->addUniform(Registry::shaderFactory()->createUniformForGLMode(
-            GL_LIGHTING, 
+            GL_LIGHTING,
             _mapNodeOptions.enableLighting().value() ? 1 : 0));
 
-        stateset->setMode( 
-            GL_LIGHTING, 
+        stateset->setMode(
+            GL_LIGHTING,
             _mapNodeOptions.enableLighting().value() ? 1 : 0);
     }
 
@@ -366,18 +388,27 @@ MapNode::init()
         OE_INFO << LC << "Adding ellipsoid uniforms.\n";
 
         // for a geocentric map, use an ellipsoid unit-frame transform and its inverse:
-        osg::Vec3d ellipFrameInverse(
-            _map->getSRS()->getEllipsoid()->getRadiusEquator(),
-            _map->getSRS()->getEllipsoid()->getRadiusEquator(),
-            _map->getSRS()->getEllipsoid()->getRadiusPolar());
-        stateset->addUniform( new osg::Uniform("oe_ellipsoidFrameInverse", osg::Vec3f(ellipFrameInverse)) );
+        if (_map->getSRS() != NULL && _map->getSRS()->getEllipsoid() != NULL)
+        {
+            osg::Vec3d ellipFrameInverse(
+                _map->getSRS()->getEllipsoid()->getRadiusEquator(),
+                _map->getSRS()->getEllipsoid()->getRadiusEquator(),
+                _map->getSRS()->getEllipsoid()->getRadiusPolar());
+            stateset->addUniform( new osg::Uniform("oe_ellipsoidFrameInverse", osg::Vec3f(ellipFrameInverse)) );
 
-        osg::Vec3d ellipFrame = osg::componentDivide(osg::Vec3d(1.0,1.0,1.0), ellipFrameInverse);
-        stateset->addUniform( new osg::Uniform("oe_ellipsoidFrame", osg::Vec3f(ellipFrame)) );
+            osg::Vec3d ellipFrame = osg::componentDivide(osg::Vec3d(1.0,1.0,1.0), ellipFrameInverse);
+            stateset->addUniform( new osg::Uniform("oe_ellipsoidFrame", osg::Vec3f(ellipFrame)) );
+        }
+        else
+        {
+            stateset->addUniform( new osg::Uniform("oe_ellipsoidFrameInverse", osg::Vec3f()) );
+            stateset->addUniform( new osg::Uniform("oe_ellipsoidFrame", osg::Vec3f()) );
+        }
     }
 
     // install the default rendermode uniform:
     stateset->addUniform( new osg::Uniform("oe_isPickCamera", false) );
+    stateset->addUniform( new osg::Uniform("oe_isShadowCamera", false) );
 
     dirtyBound();
 
@@ -394,7 +425,7 @@ MapNode::~MapNode()
 
     ModelLayerVector modelLayers;
     _map->getModelLayers( modelLayers );
-    //Remove our model callback from any of the model layers in the map    
+    //Remove our model callback from any of the model layers in the map
     for (osgEarth::ModelLayerVector::iterator itr = modelLayers.begin(); itr != modelLayers.end(); ++itr)
     {
         this->onModelLayerRemoved( itr->get() );
@@ -437,6 +468,8 @@ MapNode::getMapSRS() const
 Terrain*
 MapNode::getTerrain()
 {
+    if (getTerrainEngine() == NULL)
+          return NULL;
     return getTerrainEngine()->getTerrain();
 }
 
@@ -469,13 +502,18 @@ MapNode::addExtension(Extension* extension, const osgDB::Options* options)
         // set the IO options is they were provided:
         if ( options )
             extension->setDBOptions( options );
-        
+
+        else if ( getMap()->getReadOptions() )
+            extension->setDBOptions( getMap()->getReadOptions() );
+
         // start it.
         ExtensionInterface<MapNode>* extensionIF = ExtensionInterface<MapNode>::get(extension);
         if ( extensionIF )
         {
             extensionIF->connect( this );
         }
+
+        OE_INFO << LC << "Added extension [" << extension->getName() << "]\n";
     }
 }
 
@@ -545,27 +583,26 @@ MapNode::onModelLayerAdded( ModelLayer* layer, unsigned int index )
 {
     if ( !layer->getEnabled() )
         return;
-    
+
     // install a noe operation that will associate this mapnode with
     // any MapNodeObservers loaded by the model layer:
     ModelSource* modelSource = layer->getModelSource();
     if ( modelSource )
     {
-        // install a post-processing callback on the ModelLayer's source 
+        // install a post-processing callback on the ModelLayer's source
         // so we can update the MapNode on new data that comes in:
         modelSource->addPostMergeOperation( new MapNodeObserverInstaller(this) );
     }
 
     // create the scene graph:
-    //osg::Node* node = layer->getOrCreateSceneGraph( _map.get(), _map->getDBOptions(), 0L );
-    osg::Node* node = layer->getOrCreateSceneGraph( _map.get(), 0L );
+    osg::Node* node = layer->getOrCreateSceneGraph( _map.get(), _map->getReadOptions(), 0L );
 
     if ( node )
     {
         if ( _modelLayerNodes.find( layer ) != _modelLayerNodes.end() )
         {
             OE_WARN
-                << "Illegal: tried to add the name model layer more than once: " 
+                << "Illegal: tried to add the name model layer more than once: "
                 << layer->getName()
                 << std::endl;
         }
@@ -588,8 +625,17 @@ MapNode::onModelLayerAdded( ModelLayer* layer, unsigned int index )
                 // enfore a rendering bin if necessary:
                 if ( ms->getOptions().renderOrder().isSet() )
                 {
-                    node->getOrCreateStateSet()->setRenderBinDetails(
-                        ms->getOptions().renderOrder().value(), "RenderBin" );
+                    osg::StateSet* mss = node->getOrCreateStateSet();
+                    mss->setRenderBinDetails(
+                        ms->getOptions().renderOrder().value(),
+                        mss->getBinName().empty() ? "DepthSortedBin" : mss->getBinName());
+                }
+                if ( ms->getOptions().renderBin().isSet() )
+                {
+                    osg::StateSet* mss = node->getOrCreateStateSet();
+                    mss->setRenderBinDetails(
+                        mss->getBinNumber(),
+                        ms->getOptions().renderBin().get() );
                 }
             }
 
@@ -619,10 +665,10 @@ MapNode::onModelLayerRemoved( ModelLayer* layer )
             {
                 _models->removeChild( node );
             }
-            
+
             _modelLayerNodes.erase( i );
         }
-        
+
         dirtyBound();
     }
 }
@@ -641,7 +687,7 @@ MapNode::onModelLayerMoved( ModelLayer* layer, unsigned int oldIndex, unsigned i
             _models->removeChild( node.get() );
             _models->insertChild( newIndex, node.get() );
         }
-        
+
         dirtyBound();
     }
 }
@@ -662,7 +708,7 @@ namespace
 
 void
 MapNode::addTerrainDecorator(osg::Group* decorator)
-{    
+{
     if ( _terrainEngine )
     {
         decorator->addChild( _terrainEngine );
@@ -700,6 +746,47 @@ MapNode::removeTerrainDecorator(osg::Group* decorator)
     }
 }
 
+namespace
+{
+    template<typename T> void tryOpenLayer(T* layer)
+    {
+        if (!layer->getStatus().isError())
+        {
+            const Status& status = layer->open();
+            if (status.isError())
+            {
+                OE_WARN << LC << "Failed to open layer \"" << layer->getName() << "\" ... " << status.message() << std::endl;
+            }
+        }
+    }
+}
+
+void
+MapNode::openMapLayers()
+{
+    MapFrame frame(_map.get());
+
+    for (unsigned i = 0; i < frame.imageLayers().size(); ++i)
+    {
+        tryOpenLayer(frame.getImageLayerAt(i));
+    }
+
+    for (unsigned i = 0; i < frame.elevationLayers().size(); ++i)
+    {
+        tryOpenLayer(frame.getElevationLayerAt(i));
+    }
+
+    for (unsigned i = 0; i < frame.modelLayers().size(); ++i)
+    {
+        tryOpenLayer(frame.getModelLayerAt(i));
+    }
+
+    for (unsigned i = 0; i < frame.terrainMaskLayers().size(); ++i)
+    {
+        tryOpenLayer(frame.terrainMaskLayers().at(i).get());
+    }
+}
+
 void
 MapNode::traverse( osg::NodeVisitor& nv )
 {
@@ -714,50 +801,26 @@ MapNode::traverse( osg::NodeVisitor& nv )
             _terrainEngine->accept( v );
         }
 
+        // This is a placeholder for later, if we decide to delay the automatic opening of
+        // layers until the first traversal.
+#if 0
+        // if the model has changed, we need to queue up an update so we can open new layers.
+        if (_mapRevisionMonitor.outOfSyncWith(_map->getDataModelRevision()))
+        {
+            openMapLayers();
+            _mapRevisionMonitor.sync(_map->getDataModelRevision());
+        }
+#endif
+
         // traverse:
         std::for_each( _children.begin(), _children.end(), osg::NodeAcceptOp(nv) );
     }
 
-    else if ( nv.getVisitorType() == nv.CULL_VISITOR )
+    else if (nv.getVisitorType() == nv.UPDATE_VISITOR || nv.getVisitorType() == nv.CULL_VISITOR)
     {
-        osgUtil::CullVisitor* cv = Culling::asCullVisitor(nv);
-        if ( cv )
-        {
-            // insert traversal data for this camera:
-            osg::ref_ptr<osg::Referenced> oldUserData = cv->getUserData();
-            MapNodeCullData* cullData = getCullData( cv->getCurrentCamera() );
-            cv->setUserData( cullData );
-
-            cullData->_mapNode = this;
-
-            // calculate altitude:
-            osg::Vec3d eye = cv->getViewPoint();
-            const SpatialReference* srs = getMapSRS();
-            if ( srs && !srs->isProjected() )
-            {
-                GeoPoint lla;
-                lla.fromWorld( srs, eye );
-                cullData->_cameraAltitude = lla.alt();
-                cullData->_cameraAltitudeUniform->set( (float)lla.alt() );
-                //OE_INFO << lla.alt() << std::endl;
-            }
-            else
-            {
-                cullData->_cameraAltitude = eye.z();
-                cullData->_cameraAltitudeUniform->set( (float)eye.z() );
-            }
-
-            // window matrix:
-            cullData->_windowMatrixUniform->set( cv->getWindowMatrix() );
-
-            // traverse:
-            cv->pushStateSet( cullData->_stateSet.get() );
-            std::for_each( _children.begin(), _children.end(), osg::NodeAcceptOp(nv) );
-            cv->popStateSet();
-
-            // restore:
-            cv->setUserData( oldUserData.get() );
-        }
+        // put the MapNode in the visitor data and traverse.
+        VisitorData::store(nv, "osgEarth::MapNode", this);
+        std::for_each( _children.begin(), _children.end(), osg::NodeAcceptOp(nv) );
     }
 
     else
@@ -765,27 +828,3 @@ MapNode::traverse( osg::NodeVisitor& nv )
         osg::Group::traverse( nv );
     }
 }
-
-MapNodeCullData*
-MapNode::getCullData(osg::Camera* key) const
-{
-    // first look it up:
-    {
-        Threading::ScopedReadLock shared( _cullDataMutex );
-        CullDataMap::const_iterator i = _cullData.find( key );
-        if ( i != _cullData.end() )
-            return i->second.get();
-    }
-    // if it's not there, make it, but double-check.
-    {
-        Threading::ScopedWriteLock exclusive( _cullDataMutex );
-        
-        CullDataMap::const_iterator i = _cullData.find( key );
-        if ( i != _cullData.end() )
-            return i->second.get();
-
-        MapNodeCullData* cullData = new MapNodeCullData();
-        _cullData[key] = cullData;
-        return cullData;
-    }
-}
