@@ -1,6 +1,6 @@
 /* -*-c++-*- */
 /* osgEarth - Dynamic map generation toolkit for OpenSceneGraph
- * Copyright 2015 Pelican Mapping
+ * Copyright 2016 Pelican Mapping
  * http://osgearth.org
  *
  * osgEarth is free software; you can redistribute it and/or modify
@@ -23,6 +23,7 @@
 #include <osgEarth/Capabilities>
 #include <osgEarth/CullingUtils>
 #include <osgEarth/URI>
+#include <osgEarth/GLSLChunker>
 #include <osg/ComputeBoundsVisitor>
 #include <osgDB/FileUtils>
 #include <list>
@@ -42,6 +43,8 @@ namespace
 
     typedef std::list<const osg::StateSet*> StateSetStack;
 
+#if 0
+
     static osg::StateAttribute::GLModeValue 
     getModeValue(const StateSetStack& statesetStack, osg::StateAttribute::GLMode mode)
     {
@@ -65,7 +68,9 @@ namespace
         }
         return base_val;
     }
+#endif
     
+#if 0
     static const osg::Light*
     getLightByID(const StateSetStack& statesetStack, int id)
     {
@@ -98,6 +103,7 @@ namespace
         }
         return base_light;
     }
+#endif
     
     static const osg::Material*
     getFrontMaterial(const StateSetStack& statesetStack)
@@ -139,7 +145,7 @@ namespace
 namespace
 {
     // Code borrowed from osg::State.cpp
-    static bool replace(std::string& str, const std::string& original_phrase, const std::string& new_phrase)
+    bool replace(std::string& str, const std::string& original_phrase, const std::string& new_phrase)
     {
         bool replacedStr = false;
         std::string::size_type pos = 0;
@@ -165,26 +171,169 @@ namespace
         return replacedStr;
     }
 
-    static void replaceAndInsertDeclaration(std::string& source, std::string::size_type declPos, const std::string& originalStr, const std::string& newStr, const std::string& declarationPrefix, const std::string& declarationSuffix ="")
+    bool replaceAndInsertDeclaration(std::string& source, std::string::size_type declPos, const std::string& originalStr, const std::string& newStr, const std::string& declarationPrefix, const std::string& declarationSuffix ="")
     {
-        if (replace(source, originalStr, newStr))
+        bool yes = replace(source, originalStr, newStr);
+        if ( yes )
         {
             source.insert(declPos, declarationPrefix + newStr + declarationSuffix + std::string(";\n"));
         }
+        return yes;
     }
+
+    bool replaceAndInsertLiteral(std::string& source, std::string::size_type declPos, const std::string& originalStr, const std::string& newStr, const std::string& lit)
+    {
+        bool yes = replace(source, originalStr, newStr);
+        if ( yes )
+        {
+            source.insert(declPos, lit);
+        }
+        return yes;
+    }
+
+    int replaceVarying(GLSLChunker::Chunks& chunks, int index, const StringVector& tokens, int offset, const std::string& prefix)
+    {
+        std::stringstream buf;
+        buf << "#pragma vp_varying";
+        if ( !prefix.empty() )
+            buf << " " << prefix;
+
+        for(int i=offset; i<tokens.size(); ++i)
+        {
+            if ( !tokens[i].empty() )
+            {
+                int len = tokens[i].length();
+                if ( tokens[i].at(len-1) == ';' )
+                    buf << " " << tokens[i].substr(0, len-1); // strip semicolon
+                else
+                    buf << " " << tokens[i];
+            }
+        }
+        
+        chunks[index].text = buf.str();
+        chunks[index].type = GLSLChunker::Chunk::TYPE_DIRECTIVE;
+
+        std::stringstream buf2;
+        for(int i=offset; i<tokens.size(); ++i)
+            buf2 << (i==offset?"":" ") << tokens[i];
+
+        GLSLChunker::Chunk newChunk;
+        newChunk.type = GLSLChunker::Chunk::TYPE_STATEMENT;
+        newChunk.text = buf2.str();
+        chunks.insert( chunks.begin()+index, newChunk );
+
+        return index+1;
+    }
+
+    bool replaceVaryings(osg::Shader::Type type, GLSLChunker::Chunks& chunks)
+    {
+        bool madeChanges = false;
+
+        for(int i=0; i<chunks.size(); ++i)
+        {
+            if ( chunks[i].type == GLSLChunker::Chunk::TYPE_STATEMENT )
+            {
+                std::string replacement;
+                /*
+                StringVector tokens;
+                StringTokenizer(chunks[i].text, tokens, " \t\n", "", false, true);
+                */
+                const std::vector<std::string>& tokens = chunks[i].tokens;
+
+                if      ( tokens.size() > 1 && tokens[0] == "out" && type != osg::Shader::FRAGMENT )
+                    i = replaceVarying(chunks, i, tokens, 1, ""), madeChanges = true;
+                else if ( tokens.size() > 1 && tokens[0] == "in" && type != osg::Shader::VERTEX )
+                    i = replaceVarying(chunks, i, tokens, 1, ""), madeChanges = true;
+                else if ( tokens.size() > 2 && tokens[0] == "varying" && tokens[1] == "out" && type != osg::Shader::FRAGMENT )
+                    i = replaceVarying(chunks, i, tokens, 2, ""), madeChanges = true;
+                else if ( tokens.size() > 2 && tokens[0] == "flat" && tokens[1] == "out" && type != osg::Shader::FRAGMENT )
+                    i = replaceVarying(chunks, i, tokens, 2, "flat"), madeChanges = true;
+                else if ( tokens.size() > 2 && tokens[0] == "flat" && tokens[1] == "in" && type != osg::Shader::VERTEX )
+                    i = replaceVarying(chunks, i, tokens, 2, "flat"), madeChanges = true;
+                else if ( tokens.size() > 1 && tokens[0] == "varying" )
+                    i = replaceVarying(chunks, i, tokens, 1, ""), madeChanges = true;
+            }
+        }
+
+        return madeChanges;
+    }
+
+    void applySupportForNoFFPImpl(GLSLChunker::Chunks& chunks)
+    {
+#if !defined(OSG_GL_FIXED_FUNCTION_AVAILABLE)
+
+        // for geometry and tessellation shaders, replace the built-ins with 
+        // osg uniform aliases.
+        const char* lines[4] = {
+            "uniform mat4 osg_ModelViewMatrix;",
+            "uniform mat4 osg_ProjectionMatrix;",
+            "uniform mat4 osg_ModelViewProjectionMatrix;",
+            "uniform mat3 osg_NormalMatrix;"
+        };
+    
+        GLSLChunker chunker;
+
+        for (GLSLChunker::Chunks::iterator chunk = chunks.begin(); chunk != chunks.end(); ++chunk)
+        {
+            if (chunk->type != GLSLChunker::Chunk::TYPE_DIRECTIVE)
+            {
+                for (unsigned line = 0; line < 4; ++line) {
+                    chunk = chunks.insert(chunk, chunker.chunkLine(lines[line]));
+                    ++chunk;
+                }
+                break;
+            }
+        }
+
+        chunker.replace(chunks, "gl_ModelViewMatrix", "osg_ModelViewMatrix");
+        chunker.replace(chunks, "gl_ProjectionMatrix", "osg_ProjectionMatrix");
+        chunker.replace(chunks, "gl_ModelViewProjectionMatrix", "osg_ModelViewProjectionMatrix");
+        chunker.replace(chunks, "gl_NormalMatrix", "osg_NormalMatrix");
+    
+#endif // !defined(OSG_GL_FIXED_FUNCTION_AVAILABLE)
+    }
+}
+
+void
+ShaderPreProcessor::applySupportForNoFFP(osg::Shader* shader)
+{
+    if (!shader)
+        return;
+            
+#if !defined(OSG_GL_FIXED_FUNCTION_AVAILABLE)
+
+    GLSLChunker chunker;
+    GLSLChunker::Chunks chunks;
+    chunker.read(shader->getShaderSource(), chunks);
+
+    applySupportForNoFFPImpl(chunks);
+
+    std::string output;
+    chunker.write(chunks, output);
+    shader->setShaderSource(output);
+
+#endif // !defined(OSG_GL_FIXED_FUNCTION_AVAILABLE)
 }
 
 void
 ShaderPreProcessor::run(osg::Shader* shader)
 {
-    // only runs for non-FFP (GLES, GL3+, etc.)
-
-    if ( s_NO_FFP && shader )
+    if ( shader )
     {
+        bool dirty = false;
+
+        // only runs for non-FFP (GLES, GL3+, etc.)
         std::string source = shader->getShaderSource();
 
+        // First replace any quotes with spaces. Quotes are illegal.
+        if ( source.find('\"') != std::string::npos )
+        {
+            osgEarth::replaceIn(source, "\"", " ");
+            dirty = true;
+        }
+
         // find the first legal insertion point for replacement declarations. GLSL requires that nothing
-        // precede a "#verson" compiler directive, so we must insert new declarations after it.
+        // precede a "#version" compiler directive, so we must insert new declarations after it.
         std::string::size_type declPos = source.rfind( "#version " );
         if ( declPos != std::string::npos )
         {
@@ -197,27 +346,47 @@ ShaderPreProcessor::run(osg::Shader* shader)
             declPos = 0;
         }
 
+#if !defined(OSG_GL_FIXED_FUNCTION_AVAILABLE)
+
         int maxLights = Registry::capabilities().getMaxLights();
 
         for( int i=0; i<maxLights; ++i )
         {
-            replaceAndInsertDeclaration(
+            if ( replaceAndInsertDeclaration(
                 source, declPos,
                 Stringify() << "gl_LightSource[" << i << "]",
                 Stringify() << "osg_LightSource" << i,
                 Stringify() 
                     << osg_LightSourceParameters::glslDefinition() << "\n"
-                    << "uniform osg_LightSourceParameters " );
+                    << "uniform osg_LightSourceParameters " ) )
+            {
+                dirty = true;
+            }
 
-            replaceAndInsertDeclaration(
+            if ( replaceAndInsertDeclaration(
                 source, declPos,
                 Stringify() << "gl_FrontLightProduct[" << i << "]", 
                 Stringify() << "osg_FrontLightProduct" << i,
                 Stringify()
                     << osg_LightProducts::glslDefinition() << "\n"
-                    << "uniform osg_LightProducts " );
+                    << "uniform osg_LightProducts " ) )
+            {
+                dirty = true;
+            }
         }
+#endif
 
+        // Chunk the shader.
+        GLSLChunker chunker;
+        GLSLChunker::Chunks chunks;
+        chunker.read( source, chunks );
+
+        applySupportForNoFFPImpl(chunks);
+
+        // Replace varyings with directives that the ShaderFactory can interpret
+        // when creating interface blocks.
+        replaceVaryings( shader->getType(), chunks );
+        chunker.write( chunks, source );
         shader->setShaderSource( source );
     }
 }
@@ -405,7 +574,7 @@ UpdateLightingUniformsHelper::~UpdateLightingUniformsHelper()
 void
 UpdateLightingUniformsHelper::cullTraverse( osg::Node* node, osg::NodeVisitor* nv )
 {
-    osgUtil::CullVisitor* cv = Culling::asCullVisitor(nv);
+    osgUtil::CullVisitor* cv = static_cast<osgUtil::CullVisitor*>(nv);
     if ( cv )
     {
         StateSetStack stateSetStack;
