@@ -1,6 +1,6 @@
 /* -*-c++-*- */
 /* osgEarth - Dynamic map generation toolkit for OpenSceneGraph
- * Copyright 2015 Pelican Mapping
+ * Copyright 2016 Pelican Mapping
  * http://osgearth.org
  *
  * osgEarth is free software; you can redistribute it and/or modify
@@ -160,6 +160,8 @@ public:
         const GeoExtent&   imageExtent,
         osg::Image*        image )
     {
+        OE_DEBUG << LC << "Rendering " << features.size() << " features for " << imageExtent.toString() << "\n";
+
         // A processing context to use with the filters:
         FilterContext context( session );
         context.setProfile( getFeatureSource()->getFeatureProfile() );
@@ -482,7 +484,7 @@ class AGGLiteRasterizerTileSourceDriver : public TileSourceDriver
     public:
         AGGLiteRasterizerTileSourceDriver() {}
 
-        virtual const char* className()
+        virtual const char* className() const
         {
             return "AGG-Lite feature rasterizer";
         }
