@@ -1,6 +1,6 @@
-/* -*-c++-*- */
+/* -*_maxPolyTilingAngle_deg-c++-*- */
 /* osgEarth - Dynamic map generation toolkit for OpenSceneGraph
- * Copyright 2015 Pelican Mapping
+ * Copyright 2016 Pelican Mapping
  * http://osgearth.org
  *
  * osgEarth is free software; you can redistribute it and/or modify
@@ -26,6 +26,7 @@
 #include <osgEarthSymbology/PolygonSymbol>
 #include <osgEarthSymbology/MeshSubdivider>
 #include <osgEarthSymbology/ResourceCache>
+#include <osgEarthSymbology/MeshConsolidator>
 #include <osgEarth/Tessellator>
 #include <osgEarth/Utils>
 #include <osgEarth/Clamping>
@@ -95,7 +96,8 @@ namespace
 BuildGeometryFilter::BuildGeometryFilter( const Style& style ) :
 _style        ( style ),
 _maxAngle_deg ( 180.0 ),
-_geoInterp    ( GEOINTERP_RHUMB_LINE )
+_geoInterp    ( GEOINTERP_RHUMB_LINE ),
+_maxPolyTilingAngle_deg( 45.0f )
 {
     //nop
 }
@@ -126,8 +128,10 @@ BuildGeometryFilter::processPolygons(FeatureList& features, FilterContext& conte
             input->style().isSet() && input->style()->has<PolygonSymbol>() ? input->style()->get<PolygonSymbol>() :
             _style.get<PolygonSymbol>();
 
-        if ( !poly )
+        if ( !poly ) {
+            OE_TEST << LC << "Discarding feature with no poly symbol\n";
             continue;
+        }
 
         // run a symbol script if present.
         if ( poly->script().isSet() )
@@ -144,15 +148,17 @@ BuildGeometryFilter::processPolygons(FeatureList& features, FilterContext& conte
             part->removeDuplicates();
 
             // skip geometry that is invalid for a polygon
-            if ( part->size() < 3 )
+            if ( part->size() < 3 ) {
+                OE_TEST << LC << "Discarding illegal part (less than 3 verts)\n";
                 continue;
+            }
 
             // resolve the color:
             osg::Vec4f primaryColor = poly->fill()->color();
             
             osg::ref_ptr<osg::Geometry> osgGeom = new osg::Geometry();
-            //osgGeom->setUseVertexBufferObjects( true );
-            //osgGeom->setUseDisplayList( false );
+            osgGeom->setUseVertexBufferObjects( true );
+            osgGeom->setUseDisplayList( false );
 
             // are we embedding a feature name?
             if ( _featureNameExpr.isSet() )
@@ -177,6 +183,11 @@ BuildGeometryFilter::processPolygons(FeatureList& features, FilterContext& conte
                 l2w = _local2world;
             }
 
+            // collect all the pre-transformation HAT (Z) values.
+            osg::ref_ptr<osg::FloatArray> hats = new osg::FloatArray();
+            hats->reserve( part->size() );
+            for(Geometry::const_iterator i = part->begin(); i != part->end(); ++i )
+                hats->push_back( i->z() );
 
             // build the geometry:
             tileAndBuildPolygon(part, featureSRS, mapSRS, makeECEF, true, osgGeom, w2l);
@@ -201,7 +212,7 @@ BuildGeometryFilter::processPolygons(FeatureList& features, FilterContext& conte
                     }
 
                     double threshold = osg::DegreesToRadians( *_maxAngle_deg );
-                    OE_DEBUG << "Running mesh subdivider with threshold " << *_maxAngle_deg << std::endl;
+                    OE_TEST << "Running mesh subdivider with threshold " << *_maxAngle_deg << std::endl;
 
                     MeshSubdivider ms( _world2local, _local2world );
                     if ( input->geoInterp().isSet() )
@@ -227,13 +238,18 @@ BuildGeometryFilter::processPolygons(FeatureList& features, FilterContext& conte
                 // install clamping attributes if necessary
                 if (_style.has<AltitudeSymbol>() &&
                     _style.get<AltitudeSymbol>()->technique() == AltitudeSymbol::TECHNIQUE_GPU)
-                {            
+                {
                     Clamping::applyDefaultClampingAttrs( osgGeom, input->getDouble("__oe_verticalOffset", 0.0) );
                 }
             }
+            else
+            {
+                OE_TEST << LC << "Oh no. buildAndTilePolygon returned nothing.\n";
+            }
         }
     }
     
+    OE_TEST << LC << "Num drawables = " << geode->getNumDrawables() << "\n";
     return geode;
 }
 
@@ -296,6 +312,12 @@ BuildGeometryFilter::processPolygonizedLines(FeatureList&   features,
             if ( part->size() < 2 )
                 continue;
 
+            // collect all the pre-transformation HAT (Z) values.
+            osg::ref_ptr<osg::FloatArray> hats = new osg::FloatArray();
+            hats->reserve( part->size() );
+            for(Geometry::const_iterator i = part->begin(); i != part->end(); ++i )
+                hats->push_back( i->z() );
+
             // transform the geometry into the target SRS and localize it about 
             // a local reference point.
             osg::ref_ptr<osg::Vec3Array> verts   = new osg::Vec3Array();
@@ -318,6 +340,7 @@ BuildGeometryFilter::processPolygonizedLines(FeatureList&   features,
                 _style.get<AltitudeSymbol>()->technique() == AltitudeSymbol::TECHNIQUE_GPU)
             {
                 Clamping::applyDefaultClampingAttrs( geom, input->getDouble("__oe_verticalOffset", 0.0) );
+                Clamping::setHeights( geom, hats.get() );
             }
         }
 
@@ -372,6 +395,12 @@ BuildGeometryFilter::processLines(FeatureList& features, FilterContext& context)
             if ( part->size() < 2 )
                 continue;
 
+            // collect all the pre-transformation HAT (Z) values.
+            osg::ref_ptr<osg::FloatArray> hats = new osg::FloatArray();
+            hats->reserve( part->size() );
+            for(Geometry::const_iterator i = part->begin(); i != part->end(); ++i )
+                hats->push_back( i->z() );
+
             // if the underlying geometry is a ring (or a polygon), use a line loop; otherwise
             // use a line strip.
             GLenum primMode = dynamic_cast<Ring*>(part) ? GL_LINE_LOOP : GL_LINE_STRIP;
@@ -380,8 +409,8 @@ BuildGeometryFilter::processLines(FeatureList& features, FilterContext& context)
             osg::Vec4f primaryColor = line->stroke()->color();
             
             osg::ref_ptr<osg::Geometry> osgGeom = new osg::Geometry();
-            //osgGeom->setUseVertexBufferObjects( true );
-            //osgGeom->setUseDisplayList( false );
+            osgGeom->setUseVertexBufferObjects( true );
+            osgGeom->setUseDisplayList( false );
 
             // embed the feature name if requested. Warning: blocks geometry merge optimization!
             if ( _featureNameExpr.isSet() )
@@ -437,6 +466,7 @@ BuildGeometryFilter::processLines(FeatureList& features, FilterContext& context)
                 _style.get<AltitudeSymbol>()->technique() == AltitudeSymbol::TECHNIQUE_GPU)
             {
                 Clamping::applyDefaultClampingAttrs( osgGeom, input->getDouble("__oe_verticalOffset", 0.0) );
+                Clamping::setHeights( osgGeom, hats.get() );
             }
         }
     }
@@ -479,12 +509,18 @@ BuildGeometryFilter::processPoints(FeatureList& features, FilterContext& context
             if ( !point )
                 continue;
 
+            // collect all the pre-transformation HAT (Z) values.
+            osg::ref_ptr<osg::FloatArray> hats = new osg::FloatArray();
+            hats->reserve( part->size() );
+            for(Geometry::const_iterator i = part->begin(); i != part->end(); ++i )
+                hats->push_back( i->z() );
+
             // resolve the color:
             osg::Vec4f primaryColor = point->fill()->color();
             
             osg::ref_ptr<osg::Geometry> osgGeom = new osg::Geometry();
-            //osgGeom->setUseVertexBufferObjects( true );
-            //osgGeom->setUseDisplayList( false );
+            osgGeom->setUseVertexBufferObjects( true );
+            osgGeom->setUseDisplayList( false );
 
             // embed the feature name if requested. Warning: blocks geometry merge optimization!
             if ( _featureNameExpr.isSet() )
@@ -524,6 +560,7 @@ BuildGeometryFilter::processPoints(FeatureList& features, FilterContext& context
                 _style.get<AltitudeSymbol>()->technique() == AltitudeSymbol::TECHNIQUE_GPU)
             {            
                 Clamping::applyDefaultClampingAttrs( osgGeom, input->getDouble("__oe_verticalOffset", 0.0) );
+                Clamping::setHeights( osgGeom, hats.get() );
             }
         }
     }
@@ -531,7 +568,226 @@ BuildGeometryFilter::processPoints(FeatureList& features, FilterContext& context
     return geode;
 }
 
-#define CROP_POLYS_BEFORE_TESSELLATING 1
+// Borrowed from MeshConsolidator.cpp
+template<typename FROM, typename TO>
+osg::PrimitiveSet* copy( FROM* src, unsigned offset )
+{
+    TO* newDE = new TO( src->getMode() );
+    newDE->reserve( src->size() );
+    for( typename FROM::const_iterator i = src->begin(); i != src->end(); ++i )
+        newDE->push_back( (*i) + offset );
+    return newDE;
+}
+
+
+/**
+ * Converts an osg::Geometry to use osg::DrawElementsUInt if it doesn't already.
+ * This only works on Geometries that are already using DrawElementsUInt, DrawElementsUByte, or DrawElementsUShort
+ * We do this to normalize the primitive set types so that we can merge multiple geometries 
+ * into one later down the road.
+ */
+void convertToDrawElementsUInt(osg::Geometry* geometry)
+{
+    for (unsigned int i = 0; i < geometry->getNumPrimitiveSets(); i++)
+    {
+        osg::PrimitiveSet* ps = geometry->getPrimitiveSet(i);
+        // See if it's already a DrawElementsUInt and do nothing if it is.
+        osg::DrawElementsUInt* deUint = dynamic_cast<osg::DrawElementsUInt*>(ps);
+        if (!deUint)
+        {
+            // Copy values from the existing primitive set to a new DrawElementsUInt
+            osg::PrimitiveSet* newPS = 0;
+            if (dynamic_cast<osg::DrawElementsUByte*>(ps))
+            {
+                newPS = copy<osg::DrawElementsUByte, osg::DrawElementsUInt>(static_cast<osg::DrawElementsUByte*>(ps), 0);
+            }
+            else if (dynamic_cast<osg::DrawElementsUShort*>(ps))
+            {             
+                newPS = copy<osg::DrawElementsUShort, osg::DrawElementsUInt>(static_cast<osg::DrawElementsUShort*>(ps), 0);
+            }
+
+            // Set the new primitive set
+            if (newPS)
+            {
+                geometry->setPrimitiveSet(i, newPS);
+            }
+        }
+    }
+}
+
+/**
+ * Tesselates an osg::Geometry using the osgEarth tesselator.
+ * If it fails, fall back to the osgUtil tesselator.
+ */
+bool tesselateGeometry(osg::Geometry* geometry)
+{
+    osgEarth::Tessellator oeTess;
+    if ( !oeTess.tessellateGeometry(*geometry) )
+    {
+        osgUtil::Tessellator tess;
+        tess.setTessellationType( osgUtil::Tessellator::TESS_TYPE_GEOMETRY );
+        tess.setWindingType( osgUtil::Tessellator::TESS_WINDING_POSITIVE );
+        tess.retessellatePolygons( *geometry );
+    }
+
+    // Make sure all of the primitive sets are osg::DrawElementsUInt
+    // The osgEarth tesselator will occassionally fail, and we fall back to the osgUtil::Tesselator which can produce a mix
+    // of DrawElementsUInt, DrawElementsUByte and DrawElementsUShort depending on the number of vertices.
+    convertToDrawElementsUInt(geometry);
+    return true;
+}
+
+/**
+ * Tiles the Geometry into the given number of columns and rows
+ */
+void tileGeometry(Geometry* geometry, const SpatialReference* featureSRS, unsigned int numCols, unsigned int numRows, GeometryCollection& out)
+{
+    // Clear the output list.
+    out.clear();
+
+    Bounds b = geometry->getBounds();
+    double tw = b.width() / (double)numCols;
+    double th = b.height() / (double)numRows;
+
+    // Get the average Z, since GEOS will set teh Z of new verts to that of the cropping polygon,
+    // which is stupid but that's how it is.
+    double z = 0.0;
+    for(unsigned i=0; i<geometry->size(); ++i)
+        z += geometry->at(i).z();
+    z /= geometry->size();
+
+    osg::ref_ptr<Polygon> poly = new Polygon;
+    poly->resize( 4 );        
+
+    for(int x=0; x<(int)numCols; ++x)
+    {
+        for(int y=0; y<(int)numRows; ++y)
+        {
+            (*poly)[0].set( b.xMin() + tw*(double)x,     b.yMin() + th*(double)y,     z );
+            (*poly)[1].set( b.xMin() + tw*(double)(x+1), b.yMin() + th*(double)y,     z );
+            (*poly)[2].set( b.xMin() + tw*(double)(x+1), b.yMin() + th*(double)(y+1), z );
+            (*poly)[3].set( b.xMin() + tw*(double)x,     b.yMin() + th*(double)(y+1), z );
+
+            osg::ref_ptr<Geometry> ringTile;
+            if ( geometry->crop(poly.get(), ringTile) )
+            {
+                // Use an iterator since crop could return a multi-polygon
+                GeometryIterator gi( ringTile.get(), false );
+                while( gi.hasMore() )
+                {
+                    Geometry* geom = gi.next();
+                    out.push_back( geom );                                                
+                }
+            }
+        }
+    }
+}
+
+/**
+ * Tiles the geometry up until all the cells have less than given number of points.
+ */
+void downsizeGeometry(Geometry* geometry, const SpatialReference* featureSRS, unsigned int maxPoints, GeometryCollection& out)
+{
+    // If the geometyr is greater than the maximum number of points, we need to tile it up further.
+    if (geometry->size() > maxPoints)
+    {
+        OE_NOTICE << "Downsizing geometry of size " << geometry->size() << std::endl;
+        // Tile the geometry.
+        GeometryCollection tmp;
+        tileGeometry(geometry, featureSRS, 2, 2, tmp );
+        
+        for (unsigned int i = 0; i < tmp.size(); i++)
+        {
+            Geometry* g = tmp[i].get();
+
+            // If the generated geometry still has too many points, continue to downsample it recursively.
+            if (g->size() > maxPoints)
+            {
+                // We pass "out" as the destination here since downsizeGeometry will only append tiles that are less than the max size.
+                downsizeGeometry( g, featureSRS, maxPoints, out );
+            }
+            else
+            {
+                // Append the geometry to the output list.
+                out.push_back( g );
+            }
+        }
+    }
+    else
+    {
+        // The geometry is valid, so add it to the output list.
+        out.push_back( geometry );
+    }
+}
+
+/**
+ * Prepares a geometry into a grid if it is too big geospatially to have a sensible local tangent plane
+ * We will also tile the geometry if it just has too many points to speed up the tesselator.
+ */
+void prepareForTesselation(Geometry* geometry, const SpatialReference* featureSRS, double targetTileSizeDeg, unsigned int maxPointsPerTile, GeometryCollection& out)
+{
+    // Clear the output list.
+    GeometryCollection tiles;
+    
+    unsigned int count = geometry->size();
+
+    unsigned int tx = 1;
+    unsigned int ty = 1;
+
+    // Tile the geometry if it's geospatial size is too large to have a sensible local tangent plane.
+    GeoExtent featureExtentDeg = GeoExtent(featureSRS, geometry->getBounds()).transform(SpatialReference::create("wgs84"));
+
+    // Tile based on the extent
+    if ( featureExtentDeg.width() > targetTileSizeDeg  || featureExtentDeg.height() > targetTileSizeDeg)
+    {
+        // Determine the tile size based on the extent.
+        tx = ceil( featureExtentDeg.width() / targetTileSizeDeg );
+        ty = ceil (featureExtentDeg.height() / targetTileSizeDeg );        
+    }
+    else if (count > maxPointsPerTile)
+    {
+        // Determine the size based on the number of points.
+        unsigned numTiles = ((double)count / (double)maxPointsPerTile) + 1u;
+        tx = ceil(sqrt((double)numTiles));
+        ty = tx;        
+    }
+
+    if (tx == 1 && ty == 1)
+    {
+        // The geometry doesn't need modified so just add it to the list.
+        tiles.push_back( geometry );
+    }
+    else
+    {
+        tileGeometry( geometry, featureSRS, tx, ty, tiles );
+    }
+
+    out.clear();
+
+#if 1
+    // Just copy the output tiles to the output.
+    std::copy(tiles.begin(), tiles.end(), std::back_inserter(out));
+#else
+    // Calling this code will recursively subdivide the cells based on the number of points they have.
+    // This works but it will produces a non-regular grid which doesn't render well in geocentric
+    // due to the curvature of the earth so we disable it for now.
+    //
+    // Reduce the size of the tiles if needed.
+    for (unsigned int i = 0; i < tiles.size(); i++)
+    {
+        if (tiles[i]->size() > maxPointsPerTile)
+        {
+            GeometryCollection tmp;
+            downsizeGeometry(tiles[i].get(), featureSRS, maxPointsPerTile, tmp);
+            std::copy(tmp.begin(), tmp.end(), std::back_inserter(out));
+        }
+        else
+        {
+            out.push_back( tiles[i].get() );
+        }
+    }
+#endif
+}
 
 void
 BuildGeometryFilter::tileAndBuildPolygon(Geometry*               ring,
@@ -542,121 +798,116 @@ BuildGeometryFilter::tileAndBuildPolygon(Geometry*               ring,
                                          osg::Geometry*          osgGeom,
                                          const osg::Matrixd      &world2local)
 {
-#ifdef CROP_POLYS_BEFORE_TESSELLATING
-
 #define MAX_POINTS_PER_CROP_TILE 1024
+//#define TARGET_TILE_SIZE_EXTENT_DEGREES 5
 
-    bool built = false;
-    unsigned count = ring->getTotalPointCount();
-    if ( count > MAX_POINTS_PER_CROP_TILE )
+    if ( ring == 0L )
     {
-        unsigned tiles = (count / MAX_POINTS_PER_CROP_TILE) + 1u;
-        double tx = ceil(sqrt((double)tiles));
-        double ty = tx;
-        Bounds b = ring->getBounds();
-        double tw = b.width() / tx;
-        double th = b.height() / ty;
+        OE_WARN << LC << "Ring is NULL.\n";
+        return;
+    }
+    
+    // Tile the incoming polygon if necessary
+    // NB: this breaks down at higher latitudes; see https://github.com/gwaldron/osgearth/issues/746
 
-        OE_DEBUG << "Found " << count << " points; cropping to " << tx << " x " << ty << std::endl;
+    GeometryCollection tiles;
+    if (_maxPolyTilingAngle_deg.isSet())
+        prepareForTesselation( ring, featureSRS, _maxPolyTilingAngle_deg.get(), MAX_POINTS_PER_CROP_TILE, tiles);    
+    else
+        tiles.push_back( ring );
 
-        // Get the average Z, since GEOS will set teh Z of new verts to that of the cropping polygon,
-        // which is stupid but that's how it is.
-        double z = 0.0;
-        for(unsigned i=0; i<ring->size(); ++i)
-            z += ring->at(i).z();
-        z /= ring->size();
+    osg::ref_ptr<osg::Geode> geode = new osg::Geode;
 
-        osg::ref_ptr<Polygon> poly = new Polygon;
-        poly->resize( 4 );
+    //OE_NOTICE << LC << "TABP: tiles = " << tiles.size() << "\n";
 
-        built = true;
-        for(int x=0; x<(int)tx; ++x)
+    // Process each ring independently
+    for (int ringIndex = 0; ringIndex < tiles.size(); ringIndex++)
+    {
+        Geometry* geom = tiles[ringIndex].get();
+        if (geom)
         {
-            for(int y=0; y<(int)ty; ++y)
+            // temporary target geometry for this cell:
+            osg::ref_ptr<osg::Geometry> temp = new osg::Geometry();
+            temp->setVertexArray( new osg::Vec3Array() );
+
+            // establish a local plane for this cell based on its centroid:
+            GeoPoint cellCenter(featureSRS, geom->getBounds().center());
+            cellCenter.transform(mapSRS, cellCenter);                        
+            osg::Matrix world2cell;
+            cellCenter.createWorldToLocal( world2cell );
+
+            // build the localized polygon:
+            buildPolygon(geom, featureSRS, mapSRS, makeECEF, temp.get(), world2cell);
+
+            // if successful, transform the verts back into our master LTP:
+            if ( temp->getNumPrimitiveSets() > 0 )
             {
-                (*poly)[0].set( b.xMin() + tw*(double)x,     b.yMin() + th*(double)y,     z );
-                (*poly)[1].set( b.xMin() + tw*(double)(x+1), b.yMin() + th*(double)y,     z );
-                (*poly)[2].set( b.xMin() + tw*(double)(x+1), b.yMin() + th*(double)(y+1), z );
-                (*poly)[3].set( b.xMin() + tw*(double)x,     b.yMin() + th*(double)(y+1), z );
-                
-                osg::ref_ptr<Geometry> ringTile;
-                if ( ring->crop(poly.get(), ringTile) )
+                // Tesselate the polygon while the coordinates are still in the LTP
+                if (tesselateGeometry( temp.get() ))
                 {
-                    // Use an iterator since crop could return a multi-polygon
-                    GeometryIterator gi( ringTile.get(), false );
-                    while( gi.hasMore() )
+                    osg::Vec3Array* verts = static_cast<osg::Vec3Array*>(temp->getVertexArray());
+                    if ( verts->getNumElements() > 0 )
                     {
-                        Geometry* geom = gi.next();
-                        buildPolygon(geom, featureSRS, mapSRS, makeECEF, tessellate, osgGeom, world2local);
-                    }
-                }
-                else 
-                {
-                    // If crop resulted in empty geometry (valid case) ringTile will still be valid,
-                    // otherwise we need to process the entire polygon without tiling.
-                    if (!ringTile.valid())
-                    {
-                        //clean up geometry
-                        osgGeom->setVertexArray(0L);
-                        if (osgGeom->getNumPrimitiveSets())
-                            osgGeom->removePrimitiveSet(0, osgGeom->getNumPrimitiveSets());
+                        // Convert the coordinates back to the master LTP.
+                        // This is ok, but you will probably run into precision errors if the tile size is very large.
+                        osg::Matrix cell2world;
+                        cell2world.invert( world2cell );
+                        osg::Matrix cell2local = cell2world * world2local; // pre-multiply to avoid precision loss
 
-                        OE_NOTICE << LC << "GEOS crop failed, tessellating feature without tiling." << std::endl;
-
-                        built = false;
-                        break;
+                        for(int i=0; i<verts->size(); ++i)
+                        {
+                            (*verts)[i] = (*verts)[i] * cell2local;
+                        }
+                        geode->addDrawable(temp.get());
                     }
                 }
             }
-
-            // GEOS failed 
-            if (!built)
-                break;
         }
-    }
-
-    if ( !built )
-    {
-        buildPolygon(ring, featureSRS, mapSRS, makeECEF, tessellate, osgGeom, world2local);
-    }
-    
-
-    if ( tessellate )
-    {
-        osgEarth::Tessellator oeTess;
-        if (!oeTess.tessellateGeometry(*osgGeom))
+        else
         {
-            //fallback to osg tessellator
-            OE_DEBUG << LC << "Falling back on OSG tessellator (" << osgGeom->getName() << ")" << std::endl;
-
-            osgUtil::Tessellator tess;
-            tess.setTessellationType( osgUtil::Tessellator::TESS_TYPE_GEOMETRY );
-            tess.setWindingType( osgUtil::Tessellator::TESS_WINDING_POSITIVE );
-            tess.retessellatePolygons( *osgGeom );
+            OE_TEST << LC << "TABP: Uh oh. found a non-Ring geometry: " << geom->toString(geom->getComponentType()) << "\n";
         }
-    }    
-    
+    }
+
+    // The geode is going to contain all of our polygons now, so merge them into one.
+    osgUtil::Optimizer optimizer;
+    osgUtil::Optimizer::MergeGeometryVisitor mgv;
+    // We only want one Geometry, so don't limit the number of vertices.
+    mgv.setTargetMaximumNumberOfVertices(UINT_MAX);
+    mgv.apply( *geode.get() );
+
+    // and copy them into the output geometry.
+    if ( geode->getNumDrawables() > 0 )
+    {
+        // If we have more than one drawable after the MergeGeometryVisitor ran, we have a problem so
+        // dump out some info to help debug.
+        if (geode->getNumDrawables() != 1)
+        {
+            OE_WARN << LC << "MergeGeometryVisitor failed to merge geometries into a single one.  Num drawables " << geode->getNumDrawables() << std::endl;            
+            for (unsigned int i = 0; i < geode->getNumDrawables(); i++)
+            {
+                osg::Geometry* g = geode->getDrawable(i)->asGeometry();
+                if (g)
+                {
+                    osg::Vec3Array* verts = dynamic_cast<osg::Vec3Array*>(g->getVertexArray());
+                    if (verts)
+                    {
+                        OE_WARN << "Geometry " << i << " has " << verts->size() << " verts" << std::endl;
+                        OE_WARN << "Geometry " << i << " has " << g->getNumPrimitiveSets() << " primitive sets" << std::endl;
+                        for (unsigned int j = 0; j < g->getNumPrimitiveSets(); j++)
+                        {
+                            osg::PrimitiveSet* ps = g->getPrimitiveSet(j);
+                            OE_WARN << "PrimitiveSet " << j << ps->className() << std::endl;
+                        }
+                    }
+                }
+            }
+        }
+        osgGeom->setVertexArray( geode->getDrawable(0)->asGeometry()->getVertexArray() );
+        osgGeom->setPrimitiveSetList( geode->getDrawable(0)->asGeometry()->getPrimitiveSetList() );
+    }
+
     osgUtil::SmoothingVisitor::smooth( *osgGeom );
-
-#else
-
-    // non-cropped way
-    buildPolygon(ring, featureSRS, mapSRS, makeECEF, tessellate, osgGeom, world2local);
-    if ( tessellate )
-    {
-        osgEarth::Tessellator oeTess;
-        if (!oeTess.tessellateGeometry(*osgGeom))
-        {
-            //fallback to osg tessellator
-            OE_INFO << LC << "OE Tessellation failed! Using OSG tessellator. (" << osgGeom->getName() << ")" << std::endl;
-
-            osgUtil::Tessellator tess;
-            tess.setTessellationType( osgUtil::Tessellator::TESS_TYPE_GEOMETRY );
-            tess.setWindingType( osgUtil::Tessellator::TESS_WINDING_POSITIVE );
-            tess.retessellatePolygons( *osgGeom );
-        }
-    }
-#endif
 }
 
 // builds and tessellates a polygon (with or without holes)
@@ -665,7 +916,6 @@ BuildGeometryFilter::buildPolygon(Geometry*               ring,
                                   const SpatialReference* featureSRS,
                                   const SpatialReference* mapSRS,
                                   bool                    makeECEF,
-                                  bool                    tessellate,
                                   osg::Geometry*          osgGeom,
                                   const osg::Matrixd      &world2local)
 {
@@ -895,6 +1145,7 @@ BuildGeometryFilter::push( FeatureList& input, FilterContext& context )
         {
             switch( f->getGeometry()->getComponentType() )
             {
+            default:
             case Geometry::TYPE_LINESTRING:
             case Geometry::TYPE_RING:
                 f->style()->add( new LineSymbol() );
@@ -934,9 +1185,12 @@ BuildGeometryFilter::push( FeatureList& input, FilterContext& context )
         osg::ref_ptr<osg::Geode> geode = processPolygons(polygons, context);
         if ( geode->getNumDrawables() > 0 )
         {
+            osgUtil::Optimizer::MergeGeometryVisitor mg;
+            mg.setTargetMaximumNumberOfVertices(65536);
+            geode->accept(mg);
+
             osgUtil::Optimizer o;
             o.optimize( geode.get(), 
-                osgUtil::Optimizer::MERGE_GEOMETRY |
                 osgUtil::Optimizer::INDEX_MESH |
                 osgUtil::Optimizer::VERTEX_PRETRANSFORM |
                 osgUtil::Optimizer::VERTEX_POSTTRANSFORM );
@@ -952,9 +1206,12 @@ BuildGeometryFilter::push( FeatureList& input, FilterContext& context )
         osg::ref_ptr<osg::Geode> geode = processPolygonizedLines(polygonizedLines, twosided, context);
         if ( geode->getNumDrawables() > 0 )
         {
+            osgUtil::Optimizer::MergeGeometryVisitor mg;
+            mg.setTargetMaximumNumberOfVertices(65536);
+            geode->accept(mg);
+
             osgUtil::Optimizer o;
             o.optimize( geode.get(), 
-                osgUtil::Optimizer::MERGE_GEOMETRY |
                 osgUtil::Optimizer::INDEX_MESH |
                 osgUtil::Optimizer::VERTEX_PRETRANSFORM |
                 osgUtil::Optimizer::VERTEX_POSTTRANSFORM );
@@ -969,9 +1226,9 @@ BuildGeometryFilter::push( FeatureList& input, FilterContext& context )
         osg::ref_ptr<osg::Geode> geode = processLines(lines, context);
         if ( geode->getNumDrawables() > 0 )
         {
-            osgUtil::Optimizer o;
-            o.optimize( geode.get(), 
-                osgUtil::Optimizer::MERGE_GEOMETRY );
+            osgUtil::Optimizer::MergeGeometryVisitor mg;
+            mg.setTargetMaximumNumberOfVertices(65536);
+            geode->accept(mg);
 
             applyLineSymbology( geode->getOrCreateStateSet(), line );
             result->addChild( geode.get() );
@@ -984,9 +1241,9 @@ BuildGeometryFilter::push( FeatureList& input, FilterContext& context )
         osg::ref_ptr<osg::Geode> geode = processPoints(points, context);
         if ( geode->getNumDrawables() > 0 )
         {
-            osgUtil::Optimizer o;
-            o.optimize( geode.get(), 
-                osgUtil::Optimizer::MERGE_GEOMETRY );
+            osgUtil::Optimizer::MergeGeometryVisitor mg;
+            mg.setTargetMaximumNumberOfVertices(65536);
+            geode->accept(mg);
 
             applyPointSymbology( geode->getOrCreateStateSet(), point );
             result->addChild( geode.get() );
@@ -1014,4 +1271,4 @@ BuildGeometryFilter::push( FeatureList& input, FilterContext& context )
     {
         return 0L;
     }
-}
+}
\ No newline at end of file
