@@ -1,6 +1,6 @@
 /* -*-c++-*- */
 /* osgEarth - Dynamic map generation toolkit for OpenSceneGraph
- * Copyright 2015 Pelican Mapping
+ * Copyright 2016 Pelican Mapping
  * http://osgearth.org
  *
  * osgEarth is free software; you can redistribute it and/or modify
@@ -75,16 +75,13 @@ namespace
 //------------------------------------------------------------------------
 
 ExtrudeGeometryFilter::ExtrudeGeometryFilter() :
-_maxAngle_deg          ( 5.0 ),
 _mergeGeometry         ( true ),
 _wallAngleThresh_deg   ( 60.0 ),
 _styleDirty            ( true ),
 _makeStencilVolume     ( false ),
-_useVertexBufferObjects( true ),
-_useTextureArrays      ( true ),
 _gpuClamping           ( false )
 {
-    //NOP
+    _cosWallAngleThresh = cos( _wallAngleThresh_deg );
 }
 
 void
@@ -377,8 +374,11 @@ ExtrudeGeometryFilter::buildStructure(const Geometry*         input,
             }
 
             // transform into target SRS.
-            transformAndLocalize( corner->base, srs, corner->base, mapSRS, _world2local, makeECEF );
-            transformAndLocalize( corner->roof, srs, corner->roof, mapSRS, _world2local, makeECEF );
+            if (srs)
+            {
+                transformAndLocalize( corner->base, srs, corner->base, mapSRS, _world2local, makeECEF );
+                transformAndLocalize( corner->roof, srs, corner->roof, mapSRS, _world2local, makeECEF );
+            }
 
             // cache the length for later use.
             corner->height = (corner->roof - corner->base).length();
@@ -950,7 +950,6 @@ ExtrudeGeometryFilter::process( FeatureList& features, FilterContext& context )
             Geometry* part = iter.next();
 
             osg::ref_ptr<osg::Geometry> walls = new osg::Geometry();
-            //walls->setUseVertexBufferObjects( _useVertexBufferObjects.get() );
             
             osg::ref_ptr<osg::Geometry> rooflines = 0L;
             osg::ref_ptr<osg::Geometry> baselines = 0L;
@@ -1067,7 +1066,7 @@ ExtrudeGeometryFilter::process( FeatureList& features, FilterContext& context )
                 if ( wallSkin )
                 {
                     // Get a stateset for the individual wall stateset
-                    context.resourceCache()->getOrCreateStateSet( wallSkin, wallStateSet );
+                    context.resourceCache()->getOrCreateStateSet(wallSkin, wallStateSet, context.getDBOptions());
                 }
             }
 
@@ -1085,7 +1084,7 @@ ExtrudeGeometryFilter::process( FeatureList& features, FilterContext& context )
                 if ( roofSkin )
                 {
                     // Get a stateset for the individual roof skin
-                    context.resourceCache()->getOrCreateStateSet( roofSkin, roofStateSet );
+                    context.resourceCache()->getOrCreateStateSet(roofSkin, roofStateSet, context.getDBOptions());
                 }
             }
 
@@ -1204,26 +1203,26 @@ ExtrudeGeometryFilter::push( FeatureList& input, FilterContext& context )
 
     if ( _mergeGeometry == true && _featureNameExpr.empty() )
     {
-        osgUtil::Optimizer o;
-
-        unsigned mask = osgUtil::Optimizer::MERGE_GEOMETRY;
+        osgUtil::Optimizer::MergeGeometryVisitor mg;
+        mg.setTargetMaximumNumberOfVertices(65536);
+        group->accept(mg);
 
         // Because the mesh optimizers damaga line geometry.
         if ( !_outlineSymbol.valid() )
         {
-            mask |= osgUtil::Optimizer::INDEX_MESH;
-            mask |= osgUtil::Optimizer::VERTEX_PRETRANSFORM;
-            mask |= osgUtil::Optimizer::VERTEX_POSTTRANSFORM;
+            osgUtil::Optimizer o;
+            o.optimize(group,
+                osgUtil::Optimizer::INDEX_MESH |
+                osgUtil::Optimizer::VERTEX_PRETRANSFORM |
+                osgUtil::Optimizer::VERTEX_POSTTRANSFORM );
         }
-
-        o.optimize( group, mask );
     }
 
     // Prepare buffer objects.
     AllocateAndMergeBufferObjectsVisitor allocAndMerge;
     group->accept( allocAndMerge );
 
-    // set a uniform indiciating that clamping attributes are available.
+    // set a uniform indicating that clamping attributes are available.
     Clamping::installHasAttrsUniform( group->getOrCreateStateSet() );
 
     // if we drew outlines, apply a poly offset too.
