@@ -1,6 +1,6 @@
 /* -*-c++-*- */
 /* osgEarth - Dynamic map generation toolkit for OpenSceneGraph
-* Copyright 2015 Pelican Mapping
+* Copyright 2016 Pelican Mapping
 * http://osgearth.org
 *
 * osgEarth is free software; you can redistribute it and/or modify
@@ -32,14 +32,15 @@
 #include <osgEarthUtil/Shadowing>
 #include <osgEarthUtil/ActivityMonitorTool>
 #include <osgEarthUtil/LogarithmicDepthBuffer>
+#include <osgEarthUtil/ContourMap>
 
 #include <osgEarthUtil/LODBlending>
 #include <osgEarthUtil/VerticalScale>
-#include <osgEarthUtil/ContourMap>
 
 #include <osgEarthAnnotation/AnnotationData>
 #include <osgEarthAnnotation/AnnotationRegistry>
-#include <osgEarth/Decluttering>
+#include <osgEarth/ScreenSpaceLayout>
+#include <osgEarth/TerrainEngineNode>
 
 #include <osgEarth/XmlUtils>
 #include <osgEarth/StringUtils>
@@ -66,6 +67,8 @@ using namespace osgEarth::Symbology;
 using namespace osgEarth::Annotation;
 using namespace osgEarth::Drivers;
 
+namespace ui = osgEarth::Util::Controls;
+
 //------------------------------------------------------------------------
 
 /** Shared event handlers. */
@@ -112,6 +115,37 @@ namespace
         osg::observer_ptr<osg::Node> _node;
     };
 
+    /**
+     * Toggles the main control canvas on and off.
+     */
+    struct ToggleCanvasEventHandler : public osgGA::GUIEventHandler
+    {
+        ToggleCanvasEventHandler(osg::Node* canvas, char key) :
+            _canvas(canvas), _key(key)
+        {
+        }
+
+        bool handle( const osgGA::GUIEventAdapter& ea, osgGA::GUIActionAdapter& aa )
+        {
+            if (ea.getEventType() == osgGA::GUIEventAdapter::KEYDOWN)
+            {
+                if (ea.getKey() == _key)
+                {
+                    osg::ref_ptr< osg::Node > safeNode = _canvas.get();
+                    if (safeNode.valid())
+                    {
+                        safeNode->setNodeMask( safeNode->getNodeMask() ? 0 : ~0 );
+                    }
+                    return true;
+                }
+            }
+            return false;
+        }
+
+        osg::observer_ptr<osg::Node> _canvas;
+        char _key;
+    };
+
     // sets a user-specified uniform.
     struct ApplyValueUniform : public ControlEventHandler
     {
@@ -145,152 +179,6 @@ MouseCoordsControlFactory::create(MapNode*         mapNode,
 
 //------------------------------------------------------------------------
 
-namespace
-{
-    struct SkyTimeSliderHandler : public ControlEventHandler
-    {
-        SkyTimeSliderHandler(SkyNode* sky) : _sky(sky)  { }
-
-        SkyNode* _sky;
-
-        virtual void onValueChanged( class Control* control, float value )
-        {
-            DateTime d = _sky->getDateTime();
-            _sky->setDateTime(DateTime(d.year(), d.month(), d.day(), value));
-        }
-    };
-
-//#undef USE_AMBIENT_SLIDER
-#define USE_AMBIENT_SLIDER 1
-
-#ifdef USE_AMBIENT_SLIDER
-    struct AmbientBrightnessHandler : public ControlEventHandler
-    {
-        AmbientBrightnessHandler(SkyNode* sky) : _sky(sky) { }
-
-        SkyNode* _sky;
-
-        virtual void onValueChanged( class Control* control, float value )
-        {
-            _sky->setMinimumAmbient(osg::Vec4(value,value,value,1));
-        }
-    };
-#endif
-
-    struct AnimateSkyUpdateCallback : public osg::NodeCallback
-    {    
-        /**
-        * Creates an AnimateSkyCallback.  
-        * @param rate    The time multipler from real time.  Default of 1440 means 1 minute real time will equal 1 day simulation time.
-        */
-        AnimateSkyUpdateCallback( double rate = 1440 ):
-            _rate( rate ),
-            _prevTime( -1 ),
-            _accumTime( 0.0 )
-        {
-        }
-
-        virtual void operator()(osg::Node* node, osg::NodeVisitor* nv)
-        {             
-            SkyNode* sky = dynamic_cast< SkyNode* >( node );
-            if (sky)
-            {            
-                double time = nv->getFrameStamp()->getSimulationTime();            
-                if (_prevTime > 0)
-                {                
-                    TimeStamp t = sky->getDateTime().asTimeStamp();                  
-                    double delta = ceil((time - _prevTime) * _rate);
-                    _accumTime += delta;
-                    // The time stamp only works in seconds so we wait until we've accumulated at least 1 second to change the date.
-                    if (_accumTime > 1.0)
-                    {
-                        double deltaS = floor(_accumTime );                    
-                        _accumTime -= deltaS;
-                        t += deltaS;
-                        sky->setDateTime( t );                        
-                    }                
-                }            
-                _prevTime = time;
-            }
-            traverse( node, nv );
-        }
-
-        double _accumTime;
-        double _prevTime;    
-        double _rate;
-    };
-
-}
-
-Control*
-SkyControlFactory::create(SkyNode*         sky,
-                          osgViewer::View* view) const
-{
-    Grid* grid = new Grid();
-    grid->setChildVertAlign( Control::ALIGN_CENTER );
-    grid->setChildSpacing( 10 );
-    grid->setHorizFill( true );
-
-    grid->setControl( 0, 0, new LabelControl("Time (Hours UTC): ", 16) );
-
-    DateTime dt = sky->getDateTime();
-
-    HSliderControl* skySlider = grid->setControl(1, 0, new HSliderControl( 0.0f, 24.0f, dt.hours() ));
-    skySlider->setHorizFill( true, 300 );
-    skySlider->addEventHandler( new SkyTimeSliderHandler(sky) );
-
-    grid->setControl(2, 0, new LabelControl(skySlider) );
-
-#ifdef USE_AMBIENT_SLIDER
-    grid->setControl(0, 1, new LabelControl("Min.Ambient: ", 16) );
-    HSliderControl* ambient = grid->setControl(1, 1, new HSliderControl(0.0f, 1.0f, sky->getSunLight()->getAmbient().r()));
-    ambient->addEventHandler( new AmbientBrightnessHandler(sky) );
-    grid->setControl(2, 1, new LabelControl(ambient) );
-#endif
-
-    return grid;
-}
-
-//------------------------------------------------------------------------
-
-namespace
-{
-    struct ChangeSeaLevel : public ControlEventHandler
-    {
-        ChangeSeaLevel( OceanNode* ocean ) : _ocean(ocean) { }
-
-        OceanNode* _ocean;
-
-        virtual void onValueChanged( class Control* control, float value )
-        {
-            _ocean->setSeaLevel( value );
-        }
-    };
-}
-
-Control*
-OceanControlFactory::create(OceanNode* ocean) const
-{
-    VBox* main = new VBox();
-
-    HBox* oceanBox1 = main->addControl(new HBox());
-    oceanBox1->setChildVertAlign( Control::ALIGN_CENTER );
-    oceanBox1->setChildSpacing( 10 );
-    oceanBox1->setHorizFill( true );
-
-    oceanBox1->addControl( new LabelControl("Sea Level: ", 16) );
-
-    HSliderControl* mslSlider = oceanBox1->addControl(new HSliderControl( -250.0f, 250.0f, 0.0f ));
-    mslSlider->setBackColor( Color::Gray );
-    mslSlider->setHeight( 12 );
-    mslSlider->setHorizFill( true, 200 );
-    mslSlider->addEventHandler( new ChangeSeaLevel(ocean) );
-
-    return main;
-}
-
-//------------------------------------------------------------------------
-
 namespace
 {
     struct AnnoControlBuilder : public osg::NodeVisitor
@@ -359,17 +247,17 @@ AnnotationGraphControlFactory::create(osg::Node*       graph,
 #define LC "[MapNodeHelper] "
 
 osg::Group*
-MapNodeHelper::load(osg::ArgumentParser& args,
-                    osgViewer::View*     view,
-                    Container*           userContainer ) const
+MapNodeHelper::load(osg::ArgumentParser&  args,
+                    osgViewer::View*      view,
+                    Container*            userContainer,
+                    const osgDB::Options* readOptions) const
 {
     // do this first before scanning for an earth file
     std::string outEarth;
     args.read( "--out-earth", outEarth );
 
-    osg::ref_ptr<osgDB::Options> options = new osgDB::Options();
+    osg::ref_ptr<osgDB::Options> myReadOptions = Registry::cloneOrCreateOptions(readOptions);
     
-#if 1
     Config c;
     c.add("elevation_smoothing", false);
     TerrainOptions to(c);
@@ -377,25 +265,15 @@ MapNodeHelper::load(osg::ArgumentParser& args,
     MapNodeOptions defMNO;
     defMNO.setTerrainOptions( to );
 
-    options->setPluginStringData("osgEarth.defaultOptions", defMNO.getConfig().toJSON());
-#endif
+    myReadOptions->setPluginStringData("osgEarth.defaultOptions", defMNO.getConfig().toJSON());
 
     // read in the Earth file:
-    osg::Node* node = 0L;
-    for( int i=0; i<args.argc(); ++i )
-    {
-        if ( osgDB::getLowerCaseFileExtension(args[i]) == "earth" )
-        {
-            node = osgDB::readNodeFile( args[i], options );
-            args.remove(i);
-            break;
-        }
-    }
+    osg::Node* node = osgDB::readNodeFiles(args, myReadOptions.get());
 
     osg::ref_ptr<MapNode> mapNode;
     if ( !node )
     {
-        if ( !args.find("--images") )
+        if ( args.find("--images") < 0 )
         {
             OE_WARN << LC << "No earth file." << std::endl;
             return 0L;
@@ -428,7 +306,7 @@ MapNodeHelper::load(osg::ArgumentParser& args,
     // a root node to hold everything:
     osg::Group* root = new osg::Group();
     
-    root->addChild( mapNode.get() );
+    root->addChild( node );
 
     // parses common cmdline arguments.
     if ( view )
@@ -484,30 +362,16 @@ MapNodeHelper::parse(MapNode*             mapNode,
     osg::ref_ptr<osgDB::Options> dbOptions = Registry::instance()->cloneOrCreateOptions();
 
     // parse out custom example arguments first:
-    bool useSky        = args.read("--sky");
-    bool useOcean      = args.read("--ocean");
     bool useMGRS       = args.read("--mgrs");
     bool useDMS        = args.read("--dms");
     bool useDD         = args.read("--dd");
     bool useCoords     = args.read("--coords") || useMGRS || useDMS || useDD;
-    bool useOrtho      = args.read("--ortho");
+
     bool useAutoClip   = args.read("--autoclip");
-    bool useShadows    = args.read("--shadows");
-    bool animateSky    = args.read("--animate-sky");
     bool showActivity  = args.read("--activity");
     bool useLogDepth   = args.read("--logdepth");
     bool useLogDepth2  = args.read("--logdepth2");
     bool kmlUI         = args.read("--kmlui");
-    bool inspect       = args.read("--inspect");
-
-    if (args.read("--verbose"))
-        osgEarth::setNotifyLevel(osg::INFO);
-    
-    if (args.read("--quiet"))
-        osgEarth::setNotifyLevel(osg::FATAL);
-
-    float ambientBrightness = 0.2f;
-    args.read("--ambientBrightness", ambientBrightness);
 
     std::string kmlFile;
     args.read( "--kml", kmlFile );
@@ -543,92 +407,15 @@ MapNodeHelper::parse(MapNode*             mapNode,
     }
     canvas->addControl( mainContainer );
 
+    // Add an event handler to toggle the canvas with a key press;
+    view->addEventHandler(new ToggleCanvasEventHandler(canvas, 'y'));
 
     // look for external data in the map node:
     const Config& externals = mapNode->externalConfig();
 
-    const Config& skyConf         = externals.child("sky");
-    const Config& oceanConf       = externals.child("ocean");
-    const Config& annoConf        = externals.child("annotations");
-    const Config& declutterConf   = externals.child("decluttering");
-
     // some terrain effects.
     // TODO: Most of these are likely to move into extensions.
-    const Config& lodBlendingConf = externals.child("lod_blending");
     const Config& vertScaleConf   = externals.child("vertical_scale");
-    const Config& contourMapConf  = externals.child("contour_map");
-
-    // Adding a sky model:
-    if ( useSky || !skyConf.empty() )
-    {
-        SkyOptions options(skyConf);
-        if ( options.getDriver().empty() )
-        {
-            if ( mapNode->getMapSRS()->isGeographic() )
-                options.setDriver("simple");
-            else
-                options.setDriver("gl");
-        }
-
-        SkyNode* sky = SkyNode::create(options, mapNode);
-        if ( sky )
-        {
-            sky->attach( view, 0 );
-            if ( mapNode->getNumParents() > 0 )
-            {
-                osgEarth::insertGroup(sky, mapNode->getParent(0));
-            }
-            else
-            {
-                sky->addChild( mapNode );
-                root = sky;
-            }
-                
-            Control* c = SkyControlFactory().create(sky, view);
-            if ( c )
-                mainContainer->addControl( c );
-
-            if (animateSky)
-            {
-                sky->setUpdateCallback( new AnimateSkyUpdateCallback() );
-            }
-
-        }
-    }
-
-    // Adding an ocean model:
-    if ( useOcean || !oceanConf.empty() )
-    {
-        OceanNode* ocean = OceanNode::create(OceanOptions(oceanConf), mapNode);
-        if ( ocean )
-        {
-            // if there's a sky, we want to ocean under it
-            osg::Group* parent = osgEarth::findTopMostNodeOfType<SkyNode>(root);
-            if ( !parent ) parent = root;
-            parent->addChild( ocean );
-
-            Control* c = OceanControlFactory().create(ocean);
-            if ( c )
-                mainContainer->addControl(c);
-        }
-    }
-
-    // Shadowing.
-    if ( useShadows )
-    {
-        ShadowCaster* caster = new ShadowCaster();
-        caster->setLight( view->getLight() );
-        caster->getShadowCastingGroup()->addChild( mapNode->getModelLayerGroup() );
-        if ( mapNode->getNumParents() > 0 )
-        {
-            insertGroup(caster, mapNode->getParent(0));
-        }
-        else
-        {
-            caster->addChild(mapNode);
-            root = caster;
-        }
-    }
 
     // Loading KML from the command line:
     if ( !kmlFile.empty() )
@@ -641,6 +428,10 @@ MapNodeHelper::parse(MapNode*             mapNode,
         defaultIcon->url()->setLiteral(KML_PUSHPIN_URL);
         kml_options.defaultIconSymbol() = defaultIcon;
 
+        TextSymbol* defaultText = new TextSymbol();
+        defaultText->halo() = Stroke(0.3,0.3,0.3,1.0);
+        kml_options.defaultTextSymbol() = defaultText;
+
         osg::Node* kml = KML::load( URI(kmlFile), mapNode, kml_options );
         if ( kml )
         {
@@ -661,23 +452,11 @@ MapNodeHelper::parse(MapNode*             mapNode,
         }
     }
 
-    // Annotations in the map node externals:
-    if ( !annoConf.empty() )
-    {
-        osg::Group* annotations = 0L;
-        AnnotationRegistry::instance()->create( mapNode, annoConf, dbOptions.get(), annotations );
-        if ( annotations )
-        {
-            mapNode->addChild( annotations );
-            //root->addChild( annotations );
-        }
-    }
-
-    // Configure the de-cluttering engine for labels and annotations:
-    if ( !declutterConf.empty() )
-    {
-        Decluttering::setOptions( DeclutteringOptions(declutterConf) );
-    }
+    //// Configure the de-cluttering engine for labels and annotations:
+    //if ( !screenSpaceLayoutConf.empty() )
+    //{
+    //    ScreenSpaceLayout::setOptions( ScreenSpaceLayoutOptions(screenSpaceLayoutConf) );
+    //}
 
     // Configure the mouse coordinate readout:
     if ( useCoords )
@@ -701,14 +480,9 @@ MapNodeHelper::parse(MapNode*             mapNode,
     }
 
     // Configure for an ortho camera:
-    if ( useOrtho )
+    if ( args.read("--ortho") )
     {
         view->getCamera()->setProjectionMatrixAsOrtho(-1, 1, -1, 1, 0, 1);
-        //EarthManipulator* manip = dynamic_cast<EarthManipulator*>(view->getCameraManipulator());
-        //if ( manip )
-        //{
-        //    manip->getSettings()->setCameraProjection( EarthManipulator::PROJ_ORTHOGRAPHIC );
-        //}
     }
 
     // activity monitor (debugging)
@@ -771,22 +545,22 @@ MapNodeHelper::parse(MapNode*             mapNode,
         OE_INFO << LC << "...found " << imageLayers.size() << " image layers." << std::endl;
     }
 
-    // Install elevation morphing
-    if ( !lodBlendingConf.empty() )
-    {
-        mapNode->getTerrainEngine()->addEffect( new LODBlending(lodBlendingConf) );
-    }
-
-    // Install vertical scaler
+    // Install vertical scaler.
+    // TODO: deprecate this, or move it to an extension.
     if ( !vertScaleConf.empty() )
     {
         mapNode->getTerrainEngine()->addEffect( new VerticalScale(vertScaleConf) );
     }
 
     // Install a contour map effect.
-    if ( !contourMapConf.empty() )
+    if (args.read("--contourmap"))
     {
-        mapNode->getTerrainEngine()->addEffect( new ContourMap(contourMapConf) );
+        mapNode->addExtension(Extension::create("contourmap", ConfigOptions()));
+
+        // with the cmdline switch, hids all the image layer so we can see the contour map.
+        for (unsigned i = 0; i < mapNode->getMap()->getNumImageLayers(); ++i) {
+            mapNode->getMap()->getImageLayerAt(i)->setVisible(false);
+        }
     }
 
     // Generic named value uniform with min/max.
@@ -817,13 +591,41 @@ MapNodeHelper::parse(MapNode*             mapNode,
         }
     }
 
-    if ( inspect )
+    // Map inspector:
+    if (args.read("--inspect"))
     {
         mapNode->addExtension( Extension::create("mapinspector", ConfigOptions()) );
     }
+
+    // Memory monitor:
+    if (args.read("--monitor"))
+    {
+        mapNode->addExtension(Extension::create("monitor", ConfigOptions()) );
+    }
+
+    // Simple sky model:
+    if (args.read("--sky"))
+    {
+        mapNode->addExtension(Extension::create("sky_simple", ConfigOptions()) );
+    }
+
+    // Simple ocean model:
+    if (args.read("--ocean"))
+    {
+        mapNode->addExtension(Extension::create("ocean_simple", ConfigOptions()));
+    }
+
+    // Arbitrary extension:
+    std::string extname;
+    if (args.read("--extension", extname))
+    {
+        Extension* ext = Extension::create(extname, ConfigOptions());
+        if (ext)
+            mapNode->addExtension(ext);
+    }
     
 
-    // Process extensions.
+    // Hook up the extensions!
     for(std::vector<osg::ref_ptr<Extension> >::const_iterator eiter = mapNode->getExtensions().begin();
         eiter != mapNode->getExtensions().end();
         ++eiter)
@@ -841,6 +643,32 @@ MapNodeHelper::parse(MapNode*             mapNode,
             controlIF->connect( mainContainer );
     }
 
+
+    // Shadowing. This is last because it needs access to a light which may be provided
+    // by one of the Sky extensions.
+    if (args.read("--shadows"))
+    {
+        int unit;
+        if ( mapNode->getTerrainEngine()->getResources()->reserveTextureImageUnit(unit, "ShadowCaster") )
+        {
+            ShadowCaster* caster = new ShadowCaster();
+            caster->setTextureImageUnit( unit );
+            caster->setLight( view->getLight() );
+            caster->getShadowCastingGroup()->addChild( mapNode->getModelLayerGroup() );
+            //insertParent(caster, mapNode);
+            //root = findTopOfGraph(caster)->asGroup();
+            if ( mapNode->getNumParents() > 0 )
+            {
+                insertGroup(caster, mapNode->getParent(0));
+            }
+            else
+            {
+                caster->addChild(mapNode);
+                root = caster;
+            }
+        }
+    }
+
     root->addChild( canvas );
 }
 
@@ -855,6 +683,7 @@ MapNodeHelper::configureView( osgViewer::View* view ) const
     view->addEventHandler(new osgViewer::LODScaleHandler());
     view->addEventHandler(new osgGA::StateSetManipulator(view->getCamera()->getOrCreateStateSet()));
     view->addEventHandler(new osgViewer::RecordCameraPathHandler());
+    view->addEventHandler(new osgViewer::ScreenCaptureHandler());
 }
 
 
@@ -872,9 +701,112 @@ MapNodeHelper::usage() const
         << "  --ortho                       : use an orthographic camera\n"
         << "  --logdepth                    : activates the logarithmic depth buffer\n"
         << "  --logdepth2                   : activates logarithmic depth buffer with per-fragment interpolation\n"
+        << "  --shadows                     : activates model layer shadows\n"
         << "  --images [path]               : finds and loads image layers from folder [path]\n"
         << "  --image-extensions [ext,...]  : with --images, extensions to use\n"
         << "  --out-earth [file]            : write the loaded map to an earth file\n"
         << "  --uniform [name] [min] [max]  : create a uniform controller with min/max values\n"
-        << "  --path [file]                 : load and playback an animation path\n";
+        << "  --path [file]                 : load and playback an animation path\n"
+        << "  --extension [name]            : loads a named extension\n";
+}
+
+
+//........................................................................
+
+
+namespace
+{
+    struct SkyHoursSlider : public ui::ControlEventHandler
+    {
+        SkyHoursSlider(SkyNode* sky) : _sky(sky)  { }
+        SkyNode* _sky;
+        void onValueChanged(ui::Control* control, float value )
+        {
+            DateTime d = _sky->getDateTime();
+            _sky->setDateTime(DateTime(d.year(), d.month(), d.day(), value));
+        }
+    };
+    
+    static std::string s_month[12] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
+
+    struct SkyMonthSlider : public ui::ControlEventHandler
+    {
+        SkyMonthSlider(SkyNode* sky, ui::LabelControl* label) : _sky(sky), _label(label) { }
+        SkyNode* _sky;
+        ui::LabelControl* _label;
+        void onValueChanged(ui::Control* control, float value )
+        {
+            int m = std::min((int)value, 11);
+            DateTime d = _sky->getDateTime();            
+            _sky->setDateTime(DateTime(d.year(), m, d.day(), d.hours()));
+            _label->setText(s_month[m]);
+        }
+    };
+
+    struct SkyYearSlider : public ui::ControlEventHandler
+    {
+        SkyYearSlider(SkyNode* sky, ui::LabelControl* label) : _sky(sky), _label(label) { }
+        SkyNode* _sky;
+        ui::LabelControl* _label;
+        void onValueChanged(ui::Control* control, float value )
+        {
+            DateTime d = _sky->getDateTime();            
+            _sky->setDateTime(DateTime((int)value, d.month(), d.day(), d.hours()));
+            _label->setText(Stringify() << (int)value);
+        }
+    };
+
+    struct AmbientBrightnessHandler : public ui::ControlEventHandler
+    {
+        AmbientBrightnessHandler(SkyNode* sky) : _sky(sky) { }
+
+        SkyNode* _sky;
+
+        void onValueChanged(ui::Control* control, float value )
+        {
+            _sky->setMinimumAmbient(osg::Vec4(value,value,value,1));
+        }
+    };
+}
+
+ui::Control* SkyControlFactory::create(SkyNode* sky)
+{
+    ui::Grid* grid = new ui::Grid();
+    grid->setChildVertAlign( ui::Control::ALIGN_CENTER );
+    grid->setChildSpacing( 10 );
+    grid->setHorizFill( true );
+
+    if (sky)
+    {
+        DateTime dt = sky->getDateTime();
+
+        int r=0;
+        grid->setControl( 0, r, new ui::LabelControl("Hours UTC: ", 16) );
+        ui::HSliderControl* skyHoursSlider = grid->setControl(1, r, new ui::HSliderControl( 0.0f, 24.0f, dt.hours() ));
+        skyHoursSlider->setHorizFill( true, 250 );
+        skyHoursSlider->addEventHandler( new SkyHoursSlider(sky) );
+        grid->setControl(2, r, new ui::LabelControl(skyHoursSlider) );
+    
+        ++r;
+        grid->setControl( 0, r, new ui::LabelControl("Month: ", 16) );
+        ui::HSliderControl* skyMonthSlider = grid->setControl(1, r, new ui::HSliderControl( 0.0f, 12.0f, dt.month() ));
+        skyMonthSlider->setHorizFill( true, 250 );
+        ui::LabelControl* monthLabel = grid->setControl(2, r, new ui::LabelControl(s_month[dt.month()]));
+        skyMonthSlider->addEventHandler( new SkyMonthSlider(sky, monthLabel) );
+    
+        ++r;
+        grid->setControl( 0, r, new ui::LabelControl("Year: ", 16) );
+        ui::HSliderControl* skyYearSlider = grid->setControl(1, r, new ui::HSliderControl( 1970.0f, 2061.0f, dt.year() ));
+        skyYearSlider->setHorizFill( true, 250 );
+        ui::LabelControl* yearLabel = grid->setControl(2, r, new ui::LabelControl(Stringify()<<dt.year()));
+        skyYearSlider->addEventHandler( new SkyYearSlider(sky, yearLabel) );
+
+        ++r;
+        grid->setControl(0, r, new ui::LabelControl("Min.Ambient: ", 16) );
+        ui::HSliderControl* ambient = grid->setControl(1, r, new ui::HSliderControl(0.0f, 1.0f, sky->getSunLight()->getAmbient().r()));
+        ambient->addEventHandler( new AmbientBrightnessHandler(sky) );
+        grid->setControl(2, r, new ui::LabelControl(ambient) );
+    }
+
+    return grid;
 }
